#region License
// (c) Garry McGlennon 2009
// 
// This project is made available under the Microsoft Public License (Ms-PL)
// Code is provided as is and with no warrenty  Use at your own risk
// 
// View the project and the latest code at http://dnpframework.codeplex.com/
// 
// All other rights reserved.
// 
// blog: www.dotNetProfessional.com/blog/
#endregion
using dnp.Framework.Database.DatabaseExplorer;
using dnp.DomainModeler;
using dnp.DomainModeler.Configuration;
using dnp.Framework.Reflection;
using MbUnit.Framework;

namespace dnp.DomainModeler.Test
{
    [TestFixture]
    public class EntityTest
    {
        private Domain _domain = null;

        [TestFixtureSetUp]
        public void Setup()
        {
            DomainModelDefinition def =
                Serializer.DeserializeFromResource<DomainModelDefinition>("ReferenceSchema-DomainDefinition.xml");
            _domain = new Domain(def);
        }

        /// <summary>
        /// Ensure that entities with no base class are reported correctly
        /// </summary>
        [Test]
        public void BaseClass_None()
        {
            // The following shouldn't have a base class
            Entity entity = _domain.Entities["Car"];
            Assert.IsNull(entity.BaseClass, "Entity.BaseClass");

            entity = _domain.Entities["Person"];
            Assert.IsNull(entity.BaseClass, "Entity.BaseClass");

            entity = _domain.Entities["SingleTable"];
            Assert.IsNull(entity.BaseClass, "Entity.BaseClass");
        }

        /// <summary>
        /// Ensure that entities with a base class are reported correctly
        /// </summary>
        [Test]
        public void BaseClass_HasBaseClass()
        {
            // The following should have a base class
            Entity entity = _domain.Entities["Student"];
            Assert.AreEqual(_domain.Entities["Person"], entity.BaseClass, "Entity.BaseClass");

            entity = _domain.Entities["PreviousUniversityStudent"];
            Assert.AreEqual(_domain.Entities["Student"], entity.BaseClass, "Entity.BaseClass");
        }

        /// <summary>
        /// Ensure that entities with a base class are reported correctly
        /// </summary>
        [Test]
        public void BaseClasses_HasBaseClasses()
        {
            // The following should have a base class
            Entity entity = _domain.Entities["Student"];
            Assert.AreEqual(1, entity.BaseClasses.Count, "Entity.BaseClasses.Count");
            Assert.AreEqual(_domain.Entities["Person"], entity.BaseClasses[0], "Entity.BaseClasses");

            entity = _domain.Entities["PreviousUniversityStudent"];
            Assert.AreEqual(2, entity.BaseClasses.Count, "Entity.BaseClasses.Count");
            Assert.AreEqual(_domain.Entities["Student"], entity.BaseClasses[0], "Entity.BaseClasses");
            Assert.AreEqual(_domain.Entities["Person"], entity.BaseClasses[1], "Entity.BaseClasses");
        }

        /// <summary>
        /// Ensure that entities with out a base class are reported correctly
        /// </summary>
        [Test]
        public void BaseClasses_None()
        {
            // The following should have a base class
            Entity entity = _domain.Entities["Car"];
            Assert.AreEqual(0, entity.BaseClasses.Count, "Entity.BaseClasses.Count");

            // The following should have a base class
            entity = _domain.Entities["Person"];
            Assert.AreEqual(0, entity.BaseClasses.Count, "Entity.BaseClasses.Count");

            // The following should have a base class
            entity = _domain.Entities["SingleTable"];
            Assert.AreEqual(0, entity.BaseClasses.Count, "Entity.BaseClasses.Count");
        }

        /// <summary>
        /// Ensure that entities with no base class are reported correctly
        /// </summary>
        [Test]
        public void HasBaseClass_None()
        {
            // The following shouldn't have a base class
            Entity entity = _domain.Entities["Car"];
            Assert.IsFalse(entity.HasBaseClass, "Entity.HasBaseClass");

            entity = _domain.Entities["Person"];
            Assert.IsFalse(entity.HasBaseClass, "Entity.HasBaseClass");

            entity = _domain.Entities["SingleTable"];
            Assert.IsFalse(entity.HasBaseClass, "Entity.HasBaseClass");
        }

        /// <summary>
        /// Ensure that entities with a base class are reported correctly
        /// </summary>
        [Test]
        public void HasBaseClass_HasBaseClass()
        {
            // The following should have a base class
            Entity entity = _domain.Entities["Student"];
            Assert.IsTrue(entity.HasBaseClass, "Entity.HasBaseClass");

            entity = _domain.Entities["PreviousUniversityStudent"];
            Assert.IsTrue(entity.HasBaseClass, "Entity.HasBaseClass");
        }


        [Test]
        public void DerivedClasses_None()
        {
            // The following should have a base class
            Entity entity = _domain.Entities["Car"];
            Assert.AreEqual(0, entity.DerivedClasses.Count, "Entity.DerivedClasses.Count");

            entity = _domain.Entities["PreviousUniversityStudent"];
            Assert.AreEqual(0, entity.DerivedClasses.Count, "Entity.DerivedClasses.Count");

            entity = _domain.Entities["SingleTable"];
            Assert.AreEqual(0, entity.DerivedClasses.Count, "Entity.DerivedClasses.Count");
        }

        [Test]
        public void DerivedClasses_HasDerivedClasses()
        {
            // The following should have a base class
            Entity entity = _domain.Entities["Student"];
            Assert.AreEqual(_domain.Entities["PreviousUniversityStudent"], entity.DerivedClasses[0],
                            "Entity.DerivedClasses");
            Assert.AreEqual(1, entity.DerivedClasses.Count, "Entity.DerivedClasses.Count");

            entity = _domain.Entities["Person"];
            Assert.AreEqual(2, entity.DerivedClasses.Count, "Entity.DerivedClasses.Count");
            Assert.AreEqual(_domain.Entities["Student"], entity.DerivedClasses[0], "Entity.DerivedClasses");
            Assert.AreEqual(_domain.Entities["PreviousUniversityStudent"], entity.DerivedClasses[1],
                            "Entity.DerivedClasses");
        }

        [Test]
        public void IsAbstract_True()
        {
            Assert.IsTrue(_domain.Entities["Person"].IsAbstract);
        }

        [Test]
        public void IsAbstract_False()
        {
            Assert.IsFalse(_domain.Entities["Car"].IsAbstract);
            Assert.IsFalse(_domain.Entities["PreviousUniversityStudent"].IsAbstract);
            Assert.IsFalse(_domain.Entities["Student"].IsAbstract);
        }

        [Test]
        public void IsUpdatable_True()
        {
            Assert.IsTrue(_domain.Entities["Person"].IsUpdatable);
        }

        [Test]
        public void IsUpdatable_False()
        {
            Assert.IsFalse(_domain.Entities["CourseStudent"].IsUpdatable);
        }

        [Test]
        public void RelativeNamespace()
        {
            Entity e = _domain.Entities["Car"];
            Assert.AreEqual("Vehicle", e.RelativeNamespace);

            e = _domain.Entities["Student"];
            Assert.AreEqual("", e.RelativeNamespace);
        }

        [Test]
        public void EntityName()
        {
            Entity e = _domain.Entities["Car"];
            Assert.AreEqual("Car", e.Name);

            e = _domain.Entities["Student"];
            Assert.AreEqual("Student", e.Name);
        }

        //[Test]
        //public void FullNamespace()
        //{
        //    Entity e = _domain.Entities["Car"];
        //    Assert.AreEqual("dotNetProfessional.ReferenceSchema.Business.Entities.Vehicle", e.FullNamespace);

        //    e = _domain.Entities["Student"];
        //    Assert.AreEqual("dotNetProfessional.ReferenceSchema.Business.Entities", e.FullNamespace);
        //}

        //[Test]
        //public void FullName()
        //{
        //    Entity e = _domain.Entities["Car"];
        //    Assert.AreEqual("dotNetProfessional.ReferenceSchema.Business.Entities.Vehicle.Car", e.FullName);

        //    e = _domain.Entities["Student"];
        //    Assert.AreEqual("dotNetProfessional.ReferenceSchema.Business.Entities.Student", e.FullName);
        //}

        [Test]
        public void Domain()
        {
            Assert.AreEqual(_domain, _domain.Entities[0].Domain);
        }

        [Test]
        public void Name()
        {
            // Source: Table
            Assert.AreEqual("Car", _domain.Entities["Car"].Name);
            Assert.AreEqual("Student", _domain.Entities["Student"].Name);
            Assert.AreEqual("PreviousUniversityStudent", _domain.Entities["PreviousUniversityStudent"].Name);
            Assert.AreEqual("Dealer", _domain.Entities["Dealer"].Name);

            // Source: View
            Assert.AreEqual("v_CourseStudentCount", _domain.Entities["v_CourseStudentCount"].Name);
            Assert.AreEqual("CourseStudent", _domain.Entities["CourseStudent"].Name);
        }

        [Test]
        public void Source()
        {
            // Source: Table
            Assert.AreEqual("dbo.Car", _domain.Entities["Car"].Source.FullName);
            Assert.AreEqual("dbo.Student", _domain.Entities["Student"].Source.FullName);
            Assert.AreEqual("dbo.PreviousStudent", _domain.Entities["PreviousUniversityStudent"].Source.FullName);
            Assert.AreEqual("test.Dealer", _domain.Entities["Dealer"].Source.FullName);

            // Source: View
            Assert.AreEqual("dbo.v_CourseStudentCount", _domain.Entities["v_CourseStudentCount"].Source.FullName);
            Assert.AreEqual("dbo.v_CourseStudents", _domain.Entities["CourseStudent"].Source.FullName);
        }

        [Test]
        public void BusinessKeys()
        {
            Assert.AreEqual(2, _domain.Entities["Student"].BusinessKey.Count);
            Assert.AreEqual(2, _domain.Entities["Car"].BusinessKey.Count);
        }

        [Test]
        public void Properties_Visible()
        {
            // Source: Table
            Assert.AreEqual(1, _domain.Entities["Student"].Properties().Count);
            Assert.AreEqual(6, _domain.Entities["Car"].Properties().Count);
            Assert.AreEqual(2, _domain.Entities["Person"].Properties().Count);
            Assert.AreEqual(1, _domain.Entities["Region"].Properties().Count);
 
            // Source: View
            Assert.AreEqual(4, _domain.Entities["CourseStudent"].Properties().Count);
        }


        [Test]
        public void Properties_Hidden()
        {
            Assert.AreEqual(0, _domain.Entities["Student"].Properties(VisibilityEnum.Hidden).Count);
            Assert.AreEqual(1, _domain.Entities["AllDataTypes"].Properties(VisibilityEnum.Hidden).Count);
            Assert.AreEqual(0, _domain.Entities["Person"].Properties(VisibilityEnum.Hidden).Count);
        }

        [Test]
        public void Properties_PropertyType_PrimaryKey()
        {
            Assert.AreEqual(0, _domain.Entities["AllDataTypes"].Properties(PropertyFilterEnum.PrimaryKey).Count);
            Assert.AreEqual(1, _domain.Entities["Car"].Properties(PropertyFilterEnum.PrimaryKey).Count);
            Assert.AreEqual(1, _domain.Entities["Person"].Properties(PropertyFilterEnum.PrimaryKey).Count);
            Assert.AreEqual(2, _domain.Entities["Region"].Properties(PropertyFilterEnum.PrimaryKey).Count);
        }

        [Test]
        public void Properties_PropertyType_ForeignKey()
        {
            Assert.AreEqual(0, _domain.Entities["AllDataTypes"].Properties(PropertyFilterEnum.ForeignKey).Count);
            Assert.AreEqual(4, _domain.Entities["Car"].Properties(PropertyFilterEnum.ForeignKey).Count);
            Assert.AreEqual(2, _domain.Entities["Person"].Properties(PropertyFilterEnum.ForeignKey).Count);
            Assert.AreEqual(0, _domain.Entities["Region"].Properties(PropertyFilterEnum.ForeignKey).Count);
        }

        [Test]
        public void Properties_PropertyType_Field()
        {
            Assert.AreEqual(28, _domain.Entities["AllDataTypes"].Properties(PropertyFilterEnum.Field).Count);
            Assert.AreEqual(6, _domain.Entities["Car"].Properties(PropertyFilterEnum.Field).Count);
            Assert.AreEqual(2, _domain.Entities["Person"].Properties(PropertyFilterEnum.Field).Count);
            Assert.AreEqual(1, _domain.Entities["Region"].Properties(PropertyFilterEnum.Field).Count);
        }

        [Test]
        public void Properties_All()
        {
            // Source: Table
            Assert.AreEqual(29, _domain.Entities["AllDataTypes"].Properties(PropertyFilterEnum.All, VisibilityEnum.All).Count);
            Assert.AreEqual(11, _domain.Entities["Car"].Properties(PropertyFilterEnum.All, VisibilityEnum.All).Count);
            Assert.AreEqual(5, _domain.Entities["Person"].Properties(PropertyFilterEnum.All, VisibilityEnum.All).Count);
            Assert.AreEqual(3, _domain.Entities["Region"].Properties(PropertyFilterEnum.All, VisibilityEnum.All).Count);

            // Source: View
            Assert.AreEqual(4, _domain.Entities["CourseStudent"].Properties(PropertyFilterEnum.All, VisibilityEnum.All).Count);
        }

        [Test]
        public void Relationships_All()
        {
            Assert.AreEqual(3, _domain.Entities["Student"].Relationships().Count);
            Assert.AreEqual(3, _domain.Entities["Car"].Relationships().Count);
            Assert.AreEqual(5, _domain.Entities["Person"].Relationships().Count);
        }

        [Test]
        public void Relationships_ManyToOne()
        {
            Assert.AreEqual(3, _domain.Entities["Car"].Relationships(RelationshipTypeEnum.ManyToOne).Count);
            Assert.AreEqual(0, _domain.Entities["Student"].Relationships(RelationshipTypeEnum.ManyToOne).Count);
            Assert.AreEqual(1, _domain.Entities["Person"].Relationships(RelationshipTypeEnum.ManyToOne).Count);
        }

        [Test]
        public void Relationships_OneToMany()
        {
            Assert.AreEqual(0, _domain.Entities["Student"].Relationships(RelationshipTypeEnum.OneToMany).Count);
            Assert.AreEqual(0, _domain.Entities["Car"].Relationships(RelationshipTypeEnum.OneToMany).Count);
            Assert.AreEqual(3, _domain.Entities["Person"].Relationships(RelationshipTypeEnum.OneToMany).Count);
        }

        [Test]
        public void Relationships_OneToOne()
        {
            Assert.AreEqual(1, _domain.Entities["Student"].Relationships(RelationshipTypeEnum.OneToOne).Count); 
            Assert.AreEqual(0, _domain.Entities["Car"].Relationships(RelationshipTypeEnum.OneToOne).Count); 
            Assert.AreEqual(0, _domain.Entities["Person"].Relationships(RelationshipTypeEnum.OneToOne).Count); 
        }

        [Test]
        public void Relationships_ManyToMany()
        {
            Assert.AreEqual(2, _domain.Entities["Student"].Relationships(RelationshipTypeEnum.ManyToMany).Count);
            Assert.AreEqual(1, _domain.Entities["Course"].Relationships(RelationshipTypeEnum.ManyToMany).Count);
            Assert.AreEqual(0, _domain.Entities["Person"].Relationships(RelationshipTypeEnum.ManyToMany).Count); 
        }

        [Test]
        public void Relationships_Recursive()
        {
            Assert.AreEqual(0, _domain.Entities["Student"].Relationships(RelationshipTypeEnum.Recursive).Count); 
            Assert.AreEqual(0, _domain.Entities["Car"].Relationships(RelationshipTypeEnum.Recursive).Count); 
            Assert.AreEqual(1, _domain.Entities["Person"].Relationships(RelationshipTypeEnum.Recursive).Count); 
        }

        [Test]
        public void PrimaryKey()
        {
            Assert.AreEqual(1, _domain.Entities["Student"].PrimaryKey.Properties.Count);
            Assert.AreEqual(1, _domain.Entities["Car"].PrimaryKey.Properties.Count);
            Assert.AreEqual(1, _domain.Entities["Person"].PrimaryKey.Properties.Count);
            Assert.AreEqual(2, _domain.Entities["Dealer"].PrimaryKey.Properties.Count);
        }

        [Test]
        public void EntityType()
        {
            // Source: Table
            Assert.AreEqual(EntityTypeEnum.Entity, _domain.Entities["Student"].EntityType);
            Assert.AreEqual(EntityTypeEnum.Enum, _domain.Entities["State"].EntityType);
            Assert.AreEqual(EntityTypeEnum.Reference, _domain.Entities["Region"].EntityType);
            Assert.AreEqual(EntityTypeEnum.Entity, _domain.Entities["Dealer"].EntityType);

            // Source: View
            Assert.AreEqual(EntityTypeEnum.Entity, _domain.Entities["CourseStudent"].EntityType);
        }
    }
}