#region License
// (c) Garry McGlennon 2009
// 
// This project is made available under the Microsoft Public License (Ms-PL)
// Code is provided as is and with no warrenty  Use at your own risk
// 
// View the project and the latest code at http://dnpframework.codeplex.com/
// 
// All other rights reserved.
// 
// blog: www.dotNetProfessional.com/blog/
#endregion
using System;
using dnp.DomainModeler.Configuration;
using dnp.Framework.Database.DatabaseExplorer;
using dnp.Framework.Database.InformationSchema;
using dnp.Framework.Reflection;

namespace dnp.DomainModeler
{
    public class Domain
    {
        private bool _resolveLooseRelationships;
        private EntityCollection _Entities = new EntityCollection();
        private DomainModelDefinition _definition;
        private Database _database;
        private Model _model;

        public Domain(DomainModelDefinition definition)
        {
            _definition = definition;

            this.InitializeDatabase(definition);
            this.InitializeEntities();
        }

        private void InitializeEntities()
        {
            foreach (EntityDefinition entity in this.Definition.Entities)
            {
                if (entity.SourceObjectType == "Table")
                {
                    if (this.Database.Tables.Contains(entity.SourceObjectName))
                        this.Entities.Add(new Entity(new TableSource(this.Database.Tables[entity.SourceObjectName]), this, entity));
                    else
                        throw new ArgumentException(
                            string.Format("Unable to locate the SourceObjectName {0} for the entity {1}.",
                                          entity.SourceObjectName, entity.Name));
                }

                if (entity.SourceObjectType == "View")
                {
                    if (this.Database.Views.Contains(entity.SourceObjectName))
                        this.Entities.Add(new Entity(new ViewSource(this.Database.Views[entity.SourceObjectName]), this, entity));
                    else
                        throw new ArgumentException(
                            string.Format("Unable to locate the SourceObjectName {0} for the entity {1}.",
                                          entity.SourceObjectName, entity.Name));
                }

                if (entity.SourceObjectType == "Routine")
                {
                    if (this.Database.Routines.Contains(entity.SourceObjectName))
                        this.Entities.Add(new Entity(new RoutineSource(this.Database.Routines[entity.SourceObjectName]), this, entity));
                    else
                        throw new ArgumentException(
                            string.Format("Unable to locate the SourceObjectName {0} for the entity {1}.",
                                          entity.SourceObjectName, entity.Name));
                }
            }

            // Now update their business Keys as they can only be done after all entities have been initialized.
            foreach (Entity e in this.Entities)
            {
                e.InitializeBusinessKey();
                e.InitializeBaseEntityProperties();
            }
            // Now update the mapped properties which must be done after Business Keys have been defined.
            foreach (Entity e in this.Entities)
                e.InitializeMappedProperties();

        }
        private void InitializeDatabase(DomainModelDefinition definition)
        {
            // Need to initialize the Database Explorer objects
            string assembly =
                "dnp.Framework.Database.InformationSchema.SQLServerProvider, dnp.Framework.Database.InformationSchema.SQLServerProvider";
            // Based on the information gathered create an instance of the provider.
            IInformationSchemaProvider provider = null;
            provider = ((IInformationSchemaProvider)Assembly.CreateInstanceFromString(assembly));

            // Define the system fields for this domain based on the ManyToMany rule definitions
            SystemFieldCollection systemFiels = new SystemFieldCollection();
            foreach (ManyToManyRuleDefinition r in this.Definition.ManyToManyRuleDefinition)
                systemFiels.Add(new SystemField(r.SourceObjectName));

            _database = new Database(provider, definition.ModelConnectionString, systemFiels, this.Definition.IgnoredTablesForManyToManyCalcs.Split(','));
        }

        internal DomainModelDefinition Definition
        {
            get { return _definition; }
        }


        internal Database Database
        {
            get { return _database; }
        }

        /// <summary>
        /// The name of the client who's domain is being modeled. This information can then be used in copywright notices.
        /// </summary>
        public string ClientName
        {
            get { return _definition.ClientName; }
        }

        /// <summary>
        /// This is the name of the model that is being modeled. This can be used for documentation purposes.
        /// </summary>
        public string ModelName
        {
            get { return _definition.ModelName; }
        }

        public EntityCollection Entities
        {
            get { return _Entities; }
        }

        /// <summary>
        /// If this domain is being used as part of a model, a reference to the base model will be returned.
        /// If not model has been defined, as null reference will be returned.
        /// </summary>
        public Model Model
        {
            get { return _model; }
            set { _model = value; }
        }

        #region Dependancy List
        // While the database explorer has similar (virtually identical) code to calculate the dependancies
        // it was not able to be used. This is because the Domain Model supports the addition of addition entities
        // which may not be physical tables. As a result we would easily miss entities and report incorrect results.
        // currently only table type entities support relationships, but this may change in the future.

        /// <summary>
        /// Returns an ordered list of entites based on the dependancy order. This is helpful where objects need to be
        /// created in a particular dependancy order.
        /// </summary>
        /// <param name="order">The type of ordering required</param>
        /// <returns>Ordered entities collection</returns>
        public EntityCollection EntitiesByDependancy(EntityOrderEnum order)
        {
            switch (order)
            {
                case EntityOrderEnum.InsertOrder:
                    return this.InsertOrderedEntityList();
                case EntityOrderEnum.DeleteOrder:
                    return this.DeleteOrderedEntityList();
                default:
                    break;
            }

            throw new ArgumentException("Unknown type.");
        }
        private EntityCollection InsertOrderedEntityList()
        {
            // Make a copy of the tables we have then we'll move then out
            // as we account for them

            EntityCollection unorderedEntities = new EntityCollection();
            EntityCollection orderedEntities = new EntityCollection();
            EntityCollection independantEntities = null;

            foreach (Entity e in this.Entities)
                if (e.Source.GetType() == typeof(TableSource))
                    unorderedEntities.Add(e);

            do
            {
                independantEntities = GetIndependantEntities(unorderedEntities, orderedEntities);
                // Now add the result to the ordered list and remove them from
                // the unordered list
                foreach (Entity e in independantEntities)
                {
                    orderedEntities.Add(e);
                    unorderedEntities.Remove(e);
                }
            }
            while (independantEntities.Count != 0);

            if (unorderedEntities.Count != 0)
                throw new ArgumentException(string.Format("Unable to process {0} entities. Use of this routine is therefore not recommended, please report this as a bug to the author.", unorderedEntities.Count));

            return orderedEntities;
        }

        private static EntityCollection GetIndependantEntities(EntityCollection source, EntityCollection exclusionList)
        {
            EntityCollection independantEntities = new EntityCollection();
            foreach (Entity e in source)
            {
                if (e.Relationships().Count > 0)
                {
                    bool isIndependant = true;
                    foreach (Relationship r in e.Relationships())
                    {
                        // Check if anything references this table
                        if (r.RelationshipType != RelationshipTypeEnum.OneToMany &&
                            r.RelationshipType != RelationshipTypeEnum.ManyToMany &&
                            r.RelationshipType != RelationshipTypeEnum.Recursive &&
                            exclusionList.FindBySourceObjectName(r.Source.RightConstraint.TableFullName) == null)
                        {
                            isIndependant = false;
                            break;
                        }
                    }

                    if (isIndependant)
                        independantEntities.Add(e);
                }
                else
                    // This table has zero
                    independantEntities.Add(e);
            }

            return independantEntities;
        }

        private EntityCollection DeleteOrderedEntityList()
        {
            // This is the reverse order of the insert list
            EntityCollection insertOrder = this.EntitiesByDependancy(EntityOrderEnum.InsertOrder);
            EntityCollection deleteOrder = new EntityCollection();

            for (int i = insertOrder.Count - 1; i >= 0; i--)
            {
                deleteOrder.Add(insertOrder[i]);
            }
            return deleteOrder;
        }


        #endregion
        /// <summary>
        /// Some enities have relationships which can be inferred by intermidiary relationships. When selected the system will attempt to resolve these 
        /// 'extra' relationships otherwise if false no attempt will be made to resolve these relationships.
        /// </summary>
        public bool ResolveLooseRelationships
        {
            get { return _resolveLooseRelationships; }
        }

        /// <summary>
        /// Defines the class to be used as a base (super) class for all entities. This definition is used by
        /// scripts to determine how to create entities and other related information.
        /// </summary>
        public BaseEntityClass BaseEntityClass
        {
            get { return this.Definition.BaseEntityClass; }
        }
    }
}