#region License
// (c) Garry McGlennon 2009
// 
// This project is made available under the Microsoft Public License (Ms-PL)
// Code is provided as is and with no warrenty  Use at your own risk
// 
// View the project and the latest code at http://dnpframework.codeplex.com/
// 
// All other rights reserved.
// 
// blog: www.dotNetProfessional.com/blog/
#endregion
using System;
using dnp.DomainModeler.Configuration;
using dnp.Framework.Database.DatabaseExplorer;
using dnp.Framework.Scripting;

namespace dnp.DomainModeler
{
    public class Entity
    {
        private PropertyCollection _properties = new PropertyCollection();
        private PropertyCollection _fieldProperties;
        private RelationshipCollection _relationships = new RelationshipCollection();
        private PropertyCollection _businessKey = new PropertyCollection();
        private PropertyCollection _baseEntityClassProperties = new PropertyCollection();
        private EntityDefinition _definition;
        private PrimaryKey _primaryKey;
        private EntityCollection _baseClasses;
        private EntityCollection _derivedClasses;
        private ISource _source;
        private Domain _parent;

        public Entity(ISource source, Domain parent, EntityDefinition definition)
        {
            _parent = parent;
            _source = source;

            _definition = definition;
            InitializeRelationships();
            InitializeProperties();
            InitializePrimaryKey();
            InitializeBaseClassProperties();
        }

        private void InitializeProperties()
        {
            foreach (PropertyDefinition p in this.Definition.Properties)
            {
                IColumn sourceColumn = this.Source.Columns[p.SourceObjectName];
                _properties.Add(new Property(sourceColumn, this, p));
            }
        }

        private void InitializeBaseClassProperties()
        {
            foreach (BusinessEntityAttribute p in this.Domain.BaseEntityClass.Attributes)
                if (p.AttributeRole == AttributeRoleEnum.PrimaryKey && this.PrimaryKey != null)
                {
                    // Primary keys are special as they 'may' not have the same name in every table
                    // However by defining a primary key in the business entity its assumed a single field
                    // will be used as the primary key.
                    if (this.PrimaryKey.Properties.Count == 1)
                        this.PrimaryKey.Properties[0].AttributeRole = AttributeRoleEnum.PrimaryKey;
                }
                else
                    if (this.Properties(PropertyFilterEnum.FieldPrimaryKey).Contains(p.Name))
                        this.Properties(PropertyFilterEnum.FieldPrimaryKey)[p.Name].AttributeRole = p.AttributeRole;
        }

        // This method can only be executed AFTER all entities have been initialized.
        internal void InitializeBusinessKey()
        {
            // Don't process entities that have base classes
            if (this.HasBaseClass)
                return;

            foreach (BusinessKeyDefinition k in this.Definition.BusinessKeys)
            {
                Property keyProperty = null;

                // Step 1: Identify the entity which has the business key field
                foreach (Property p in this.Properties(PropertyFilterEnum.FieldPrimaryKey))
                    if (p.Source.Name == k.SourceObjectName)
                    {
                        keyProperty = p;
                        break;
                    }

                // If we can't find the property it might be in a relationship.
                if (keyProperty == null)
                {
                    foreach (Relationship r in this.Relationships())
                    {
                        foreach (Property p in r.Properties)
                            if (p.Source.Name == k.SourceObjectName)
                            {
                                keyProperty = p;
                                break;
                            }
                        // Only continue if we've not found the property
                        if (keyProperty != null)
                            break;
                    }
                }

                if (keyProperty == null)
                    throw new ArgumentException(
                        string.Format("Unable to locate property definition for the key {0} on Entity {1}", k.SourceObjectName, this.Name));

                // Check that the business key is not computed as this isn't currently supported.
                if (keyProperty.Source.IsComputed)
                    throw new ArgumentException(string.Format("The field {0}.{1} cannot be used as a business key because it is a computed field. Please select another field for the business key.", keyProperty.Parent.Name, keyProperty.Name));

                this.BusinessKey.Add(keyProperty);
            }
        }

        #region Calculate Mapped Properties

        // This method can only be executed AFTER all entities have been initialized.
        internal void InitializeMappedProperties()
        {
            DefineMappedProperties(this);

            // Now for any property that wasn't defined as mappable we'll create a simple loop back mapping (to make scripting more consistent)
            foreach (Property p in this.Properties(PropertyFilterEnum.All))
                if (p.Mapped == null)
                    // This property hasn't been mapped to craete a pseudo mapping
                    p.Mapped = new MappedProperty(p.Name, p);
        }

        /// <summary>
        /// Deterimine if a given relationship of M:1 can be made mapable.
        /// </summary>
        /// <param name="r">M:1 relationship to test</param>
        /// <returns></returns>
        private static bool IsMappable(Relationship r)
        {
            if ((r.RelationshipType != RelationshipTypeEnum.ManyToOne && r.RelationshipType != RelationshipTypeEnum.Recursive) || r.Properties.Count != 1)
                // Only single property relationships can be mapped at this time
                return false;
            else
            {
                return true;
            }
        }

        private static void DefineMappedProperties(Entity entity)
        {
            foreach (Relationship r in entity.Relationships(RelationshipTypeEnum.Recursive | RelationshipTypeEnum.ManyToOne))
                if (IsMappable(r))
                {
                    // Ensure that any 'mappable' entity that has more than 1 property is not a Reference type
                    if (r.RelatedEntity.EntityType == EntityTypeEnum.Reference && r.RelatedEntity.BusinessKey.Count > 1)
                        throw new ArgumentException(string.Format("The entity {0} is marked as a reference type but has {1} business keys instead of 1", r.RelatedEntity.Name, r.RelatedEntity.BusinessKey.Count));

                    // Enums can only have up to two keys one for the key and one for the description
                    if (r.RelatedEntity.EntityType == EntityTypeEnum.Enum && r.RelatedEntity.BusinessKey.Count > 2)
                        throw new ArgumentException(string.Format("The entity {0} is marked as an enum type but has {1} business keys instead of 2", r.RelatedEntity.Name, r.RelatedEntity.BusinessKey.Count));

                    MappedProperty propertyMapping = new MappedProperty(r.Name);
                    // Need to cycle through each of the business keys looking for the property instance.
                    for (int keyIndex = 0; keyIndex < r.RelatedEntity.BusinessKey.Count; keyIndex++)
                    {
                        Property mappedProperty = null;
                        if (r.RelatedEntity.Properties(VisibilityEnum.Visible, PropertyDepthFilterEnum.IncludeBaseClass).Contains(r.RelatedEntity.BusinessKey[keyIndex].Name))
                            mappedProperty = r.RelatedEntity.Properties(VisibilityEnum.Visible, PropertyDepthFilterEnum.IncludeBaseClass)[r.RelatedEntity.BusinessKey[keyIndex].Name];
                        else
                        {
                            // Find the property in the relationships
                            foreach (Relationship relatedRelationship in r.RelatedEntity.Relationships(RelationshipTypeEnum.ManyToOne, PropertyDepthFilterEnum.IncludeBaseClass))
                            {
                                foreach (Property p in relatedRelationship.Properties)
                                    if (p.Source.Name == r.RelatedEntity.BusinessKey[keyIndex].Name)
                                    {
                                        mappedProperty = p;
                                        break;
                                    }
                                // Only continue if we've not found the property
                                if (mappedProperty != null)
                                    break;
                            }
                        }
                        if (mappedProperty == null)
                            throw new ArgumentException(string.Format("Unable to map property for entity {0}", entity.Name));

                        propertyMapping.Properties.Add(mappedProperty);
                    } // for keyIndex

                    // Now associate property mappings to property
                    entity.Properties(PropertyFilterEnum.All)[r.Properties[0].Name].Mapped = propertyMapping;
                } // if
        }

        #endregion
        internal void InitializeBaseEntityProperties()
        {
            PropertyCollection foundEntityProperties = new PropertyCollection();

            // Can't inherit twice so if we have a base class we have no entity base properties
            if (!this.HasBaseClass)
                foreach (BusinessEntityAttribute p in this.Domain.BaseEntityClass.Properties)
                {
                    if (this.Properties().Contains(p.Name))
                        foundEntityProperties.Add(this.Properties()[p.Name]);
                }
            // Now for this to be valid we must have the same number of properties in this entity as is
            // in the base entity. NB: If this is a derived class the base class will have the properties
            if (foundEntityProperties.Count == this.Domain.BaseEntityClass.Properties.Count)
                _baseEntityClassProperties = foundEntityProperties;
            else
                _baseEntityClassProperties = new PropertyCollection();
        }
        private void InitializeRelationships()
        {
            foreach (RelationshipDefinition r in this.Definition.Relationships)
            {
                _relationships.Add(new Relationship(this.GetSourceRelationship(r.SourceObjectName), this, r));
            }
        }

        private void InitializePrimaryKey()
        {
            if (this.Source.GetType() == typeof(TableSource))
            {
                if (((TableSource)this.Source).Instance.PrimaryKey != null)
                    _primaryKey = new PrimaryKey(((TableSource)this.Source).Instance.PrimaryKey, this, this.Parent.Definition);
            }
        }

        private Framework.Database.DatabaseExplorer.Relationship GetSourceRelationship(string name)
        {
            switch (this.Source.Instance.GetType().Name)
            {
                case "Table":
                    return ((Table)this.Source.Instance).Relationships[name];
                default:
                    throw new ArgumentException("Relationships for the the source type: " + this.Source.GetType().Name + " is currently not supported.");
            }
        }
        public EntityDefinition Definition
        {
            get { return _definition; }
        }

        public string Name
        {
            get { return _definition.Name; }
        }

        public bool Visible
        {
            get { return _definition.Visible; }
        }

        public Domain Parent
        {
            get { return _parent; }
        }
        public PrimaryKey PrimaryKey
        {
            get { return _primaryKey; }

        }
        public ISource Source
        {
            get { return _source; }
        }

        public bool IsAbstract
        {
            get { return _definition.IsAbstract; }
        }

        /// <summary>
        /// Specifies if this entity should be made updatable. Generally used to indicate if persistance should be supported.
        /// </summary>
        public bool IsUpdatable
        {
            get { return _definition.IsUpdatable; }
        }

        public string RelativeNamespace
        {
            get { return Model.FixNamespace(Naming.EvaluateNamingExpression(_definition.RelativeNamespace, this)); }
        }

        /// <summary>
        /// This returns the relative namespace + the class name. This is useful when the root domain namespace has already been defined.
        /// </summary>
        public string RelativeName
        {
            get { return Model.FixNamespace(this.RelativeNamespace + "." + this.Name); }
        }
        public string Description
        {
            get { return this.Source.Definition; }
        }
        #region Properties methods
        /// <summary>
        /// Returns all the visible properties of the entity
        /// </summary>
        /// <returns>Visible properties</returns>
        public PropertyCollection Properties()
        {
            _fieldProperties = this.Properties(PropertyFilterEnum.Field, VisibilityEnum.Visible);

            return _fieldProperties;
        }
        public PropertyCollection Properties(VisibilityEnum visibility)
        {
            return this.Properties(PropertyFilterEnum.Field, visibility);
        }
        public PropertyCollection Properties(PropertyFilterEnum filter)
        {
            return Properties(filter, VisibilityEnum.Visible);
        }


        /// <summary>
        /// Returns a collection of property objects for this entity based on the filter and visibility.
        /// </summary>
        /// <param name="propertyFilter">The type of properties to return</param>
        /// <param name="visibility">Visibility filter used to restrict properties returned</param>
        /// <returns>Filtered property collection</returns>
        public PropertyCollection Properties(PropertyFilterEnum propertyFilter, VisibilityEnum visibility)
        {
            PropertyCollection filteredProperties = new PropertyCollection();
            foreach (Property p in _properties)
            {
                // These checks are a little more complicated than usual as the property has a simple
                // boolean type but the comparision is on a flags enum.
                if ((p.AttributeRole == AttributeRoleEnum.Field ||
                    p.AttributeRole == AttributeRoleEnum.ConcurrencyKey ||
                    p.AttributeRole == AttributeRoleEnum.CreatedBy ||
                    p.AttributeRole == AttributeRoleEnum.CreatedDate ||
                    p.AttributeRole == AttributeRoleEnum.ModifiedBy ||
                    p.AttributeRole == AttributeRoleEnum.ModifiedDate) && (PropertyFilterEnum.Field == (PropertyFilterEnum.Field & propertyFilter)) && (
                    (p.Visible && (VisibilityEnum.Visible == (VisibilityEnum.Visible & visibility))) ||
                    (!p.Visible && (VisibilityEnum.Hidden == (VisibilityEnum.Hidden & visibility)))
                    ))
                    filteredProperties.Add(p);

                if (p.AttributeRole == AttributeRoleEnum.PrimaryKey && (PropertyFilterEnum.PrimaryKey == (PropertyFilterEnum.PrimaryKey & propertyFilter)) && (
                    (p.Visible && (VisibilityEnum.Visible == (VisibilityEnum.Visible & visibility))) ||
                    (!p.Visible && (VisibilityEnum.Hidden == (VisibilityEnum.Hidden & visibility)))
                    ))
                    filteredProperties.Add(p);

                if (p.AttributeRole == AttributeRoleEnum.ForeignKey && (PropertyFilterEnum.ForeignKey == (PropertyFilterEnum.ForeignKey & propertyFilter)) && (
                    (p.Visible && (VisibilityEnum.Visible == (VisibilityEnum.Visible & visibility))) ||
                    (!p.Visible && (VisibilityEnum.Hidden == (VisibilityEnum.Hidden & visibility)))
                    ))
                    filteredProperties.Add(p);
            }
            return filteredProperties;
        }
        /// <summary>
        /// Returns a collection of property objects for this entity based on the filter.
        /// Defaults property type to Field
        /// </summary>
        /// <param name="visibility">Visibility filter used to restrict properties returned</param>
        /// <param name="propertyFilter">used to determine the depth of the entity to include</param>
        /// <returns>Filtered property collection</returns>
        public PropertyCollection Properties(VisibilityEnum visibility, PropertyDepthFilterEnum propertyFilter)
        {
            return Properties(PropertyFilterEnum.All, visibility, propertyFilter);
        }

        #region Properties by View
        /// <summary>
        /// Returns a set of properties that match one of the view filters. When scripting queries the Query view is best and when dealing with code, the entity view is best.
        /// The filter is based on all visible properties including M:1 relationships.
        /// </summary>
        /// <param name="view">Choice of logical view</param>
        /// <param name="depthFilter">Determines the depth the view should include</param>
        /// <returns></returns>
        public ViewPropertyCollection Properties(PropertyViewEnum view, PropertyDepthFilterEnum depthFilter)
        {
            // Due to an issue with the base properties collection eliminating duplicate properties accross derived types
            // we have to handle dervived types independantly
            PropertyDepthFilterEnum localFilter;
            if (PropertyDepthFilterEnum.DerivedClass == (PropertyDepthFilterEnum.DerivedClass & depthFilter))
                localFilter = PropertyDepthFilterEnum.IncludeBaseClass;
            else
                localFilter = depthFilter;

            ViewPropertyCollection viewProperties = new ViewPropertyCollection();
            foreach (Property p in this.Properties(PropertyFilterEnum.All, VisibilityEnum.Visible, localFilter))
                // Ignore base class primary key for the Entity view
                if ((p.InPrimaryKey && p.Parent.Name == this.Name) || !p.InPrimaryKey || view == PropertyViewEnum.Query)
                    viewProperties.Add(p);

            if (PropertyDepthFilterEnum.DerivedClass == (PropertyDepthFilterEnum.DerivedClass & depthFilter))
                foreach (Entity de in this.DerivedClasses)
                    foreach (Property p in de.Properties(PropertyFilterEnum.All, VisibilityEnum.Visible, PropertyDepthFilterEnum.Entity))
                        // Ignore base class primary key for the Entity view
                        if ((p.InPrimaryKey && p.Parent.Name == this.Name) || !p.InPrimaryKey || view == PropertyViewEnum.Query)
                            viewProperties.Add(p);


            if (view == PropertyViewEnum.Query)
            {
                foreach (Property p in this.Properties(PropertyFilterEnum.ForeignKey | PropertyFilterEnum.PrimaryKey, VisibilityEnum.Visible, localFilter))
                {
                    if (!p.IsMapped)
                        continue;
                    // Clone the existing property (we need to do this as we are going to change it and dont want to affect the original version)
                    Property keyProperty = p.Clone() as Property;

                    keyProperty.Mapped = new MappedProperty(p.Name, p);
                    viewProperties.Add(keyProperty);
                }

                if (PropertyDepthFilterEnum.DerivedClass == (PropertyDepthFilterEnum.DerivedClass & depthFilter))
                    foreach (Entity de in this.DerivedClasses)
                        foreach (Property p in de.Properties(PropertyFilterEnum.ForeignKey | PropertyFilterEnum.PrimaryKey, VisibilityEnum.Visible, PropertyDepthFilterEnum.Entity))
                        {
                            if (!p.IsMapped)
                                continue;
                            // Clone the existing property (we need to do this as we are going to change it and dont want to affect the original version)
                            Property keyProperty = p.Clone() as Property;

                            keyProperty.Mapped = new MappedProperty(p.Name, p);
                            viewProperties.Add(keyProperty);
                        }

            }

            return viewProperties;
        }

        #endregion
        /// <summary>
        /// Returns a collection of property objects for this entity based on the filter.
        /// </summary>
        /// <param name="propertyFilter">used to determine the depth of the entity to include</param>
        /// <returns>Filtered property collection</returns>
        public PropertyCollection Properties(PropertyDepthFilterEnum propertyFilter)
        {
            return this.Properties(PropertyFilterEnum.Field, VisibilityEnum.Visible, propertyFilter);
        }

        /// <summary>
        /// Returns a collection of property objects for this entity based on the filter.
        /// </summary>
        /// <param name="propertyFilter">The type of properties to return</param>
        /// <param name="visibility">Visibility filter used to restrict properties returned</param>
        /// <param name="propertyDepthFilter">used to determine the depth of the entity to include</param>
        /// <returns>Filtered property collection</returns>
        public PropertyCollection Properties(PropertyFilterEnum propertyFilter, VisibilityEnum visibility, PropertyDepthFilterEnum propertyDepthFilter)
        {
            PropertyCollection properties = new PropertyCollection();

            if (PropertyDepthFilterEnum.Entity == (PropertyDepthFilterEnum.Entity & propertyDepthFilter))
                properties = this.Properties(propertyFilter, visibility);

            if (PropertyDepthFilterEnum.BaseClass == (PropertyDepthFilterEnum.BaseClass & propertyDepthFilter))
                foreach (Entity e in this.BaseClasses)
                    foreach (Property p in e.Properties(propertyFilter, visibility))
                        if (!properties.Contains(p.Name))
                            properties.Add(p);

            if (PropertyDepthFilterEnum.DerivedClass == (PropertyDepthFilterEnum.DerivedClass & propertyDepthFilter))
                foreach (Entity e in this.DerivedClasses)
                    foreach (Property p in e.Properties(propertyFilter, visibility))
                        if (!properties.Contains(p.Name))
                            properties.Add(p);

            return properties;
        }
        #endregion

        public RelationshipCollection Relationships(RelationshipTypeEnum filter, PropertyDepthFilterEnum propertyFilter)
        {
            RelationshipCollection relationships;

            relationships = this.Relationships(filter);
            if (PropertyDepthFilterEnum.BaseClass == (PropertyDepthFilterEnum.BaseClass & propertyFilter))
                foreach (Entity e in this.BaseClasses)
                    foreach (Relationship r in e.Relationships(filter))
                        relationships.Add(r);

            return relationships;
        }

        public RelationshipCollection Relationships()
        {
            return
                this.Relationships(RelationshipTypeEnum.ManyToMany | RelationshipTypeEnum.ManyToOne |
                                   RelationshipTypeEnum.OneToMany |
                                   RelationshipTypeEnum.OneToOne | RelationshipTypeEnum.Recursive);
        }
        public RelationshipCollection Relationships(RelationshipTypeEnum filter)
        {
            // As we have already processed the all filter previously we will just serve that back 
            if (filter == (RelationshipTypeEnum.ManyToMany |
                RelationshipTypeEnum.ManyToOne |
                RelationshipTypeEnum.OneToMany |
                RelationshipTypeEnum.OneToOne |
                RelationshipTypeEnum.Recursive))
                return _relationships;
            else
            {
                RelationshipCollection filteredRelationships =
                    new RelationshipCollection();
                foreach (Relationship r in _relationships)
                {
                    if (r.RelationshipType == (RelationshipTypeEnum.ManyToMany & filter))
                        filteredRelationships.Add(r);

                    if (r.RelationshipType == (RelationshipTypeEnum.ManyToOne & filter))
                        filteredRelationships.Add(r);

                    if (r.RelationshipType == (RelationshipTypeEnum.OneToMany & filter))
                        filteredRelationships.Add(r);

                    if (r.RelationshipType == (RelationshipTypeEnum.OneToOne & filter))
                        filteredRelationships.Add(r);

                    if (r.RelationshipType == (RelationshipTypeEnum.Recursive & filter))
                        filteredRelationships.Add(r);

                    if (r.RelationshipType == (RelationshipTypeEnum.Unknown & filter))
                        filteredRelationships.Add(r);
                }

                return filteredRelationships;
            }
        }
        /// <summary>
        /// The properties that have been defined as a business key. A business key is a unique constraint for the enity
        /// and is used primarily as a business friendly replacement for a primary key. If the entity has a base class
        /// then the business key is the one defined at the root of the entity.
        /// </summary>
        public PropertyCollection BusinessKey
        {
            get
            {
                if (this.HasBaseClass)
                    return this.RootEntity.BusinessKey;
                else
                    return _businessKey;
            }
        }

        /// <summary>
        /// Retuns if this entity supports the defined base entity properties. An entity must support all the defined
        /// properties otherwise false will be returned.
        /// </summary>
        public bool IsBaseEntityCapable
        {
            get
            {
                if (!string.IsNullOrEmpty(this.Domain.BaseEntityClass.ClassName))
                    if (this.Domain.BaseEntityClass.Properties.Count == 0)
                        return true;
                    else
                        return this.BaseEntityClassProperties.Count != 0;
                else
                    return false;
            }
        }

        public bool IsDerivedClassProperty(Property p)
        {
            if (this.DerivedClasses.Count == 0)
                // Now derived classes so must be false
                return false;

            // This property belongs to this entity
            if (p.Parent.Name == this.Name)
                return false;

            // This property belongs to one of the base classes
            foreach (Entity e in this.BaseClasses)
                if (p.Parent.Name == e.Name)
                    return false;

            return true;
        }
        /// <summary>
        /// Returns the properties that map to the user defined base entity class. These properties
        /// generally won't form part of a class structure as they are already part of the base class.
        /// </summary>
        public PropertyCollection BaseEntityClassProperties
        {
            get { return _baseEntityClassProperties; }
        }

        /// <summary>
        /// Returns the classes that inherit from this class all the way up the tree.
        /// </summary>
        public EntityCollection DerivedClasses
        {
            get
            {
                // Only do this once
                if (_derivedClasses == null)
                {
                    _derivedClasses = new EntityCollection();
                    this.GetDerivedClasses(this, _derivedClasses);
                }
                return _derivedClasses;
            }
        }

        public EntityCollection BaseClasses
        {
            get
            {
                // Only do this once
                if (_baseClasses == null)
                {
                    _baseClasses = new EntityCollection();
                    this.GetBaseClasses(this, _baseClasses);
                }
                return _baseClasses;
            }
        }

        private void GetBaseClasses(Entity entity, EntityCollection entities)
        {
            foreach (Relationship r in entity.Relationships())
            {
                if (r.RelationshipType == RelationshipTypeEnum.OneToOne)
                {
                    entities.Add(r.RelatedEntity);
                    GetBaseClasses(r.RelatedEntity, entities);
                    break;
                }
            }
        }

        private void GetDerivedClasses(Entity entity, EntityCollection entities)
        {
            foreach (Entity e in this.Domain.Entities)
            {
                if (e.HasBaseClass && e.BaseClass.Name == entity.Name)
                {
                    entities.Add(e);
                    this.GetDerivedClasses(e, entities);
                }
            }
        }

        public EntityTypeEnum EntityType
        {
            get { return _definition.EntityType; }
        }

        public bool HasBaseClass
        {
            get { return this.BaseClasses.Count > 0; }
        }

        public bool HasPrimaryKey
        {
            get { return this.PrimaryKey != null; }
        }

        public Entity BaseClass
        {
            get
            {
                if (this.HasBaseClass)
                    return this.BaseClasses[0];
                else
                    return null;
            }
        }

        public Entity RootEntity
        {
            get
            {
                if (!this.HasBaseClass)
                    return this;
                else
                {
                    Entity root = this;
                    while (root.HasBaseClass)
                        root = root.BaseClass;

                    return root;
                }
            }
        }
        public Domain Domain
        {
            get { return this.Parent; }
        }

        public override string ToString()
        {
            return this.RelativeName;
        }
    }
}
