#region License
// (c) Garry McGlennon 2009
// 
// This project is made available under the Microsoft Public License (Ms-PL)
// Code is provided as is and with no warrenty  Use at your own risk
// 
// View the project and the latest code at http://dnpframework.codeplex.com/
// 
// All other rights reserved.
// 
// blog: www.dotNetProfessional.com/blog/
#endregion
using System;
using dnp.DomainModeler.Configuration;
using dnp.Framework.Database.DatabaseExplorer;

namespace dnp.DomainModeler
{
    public class Property : DomainIdentifierBase<Entity>, ICloneable
    {
        private PropertyDefinition _definition;
        private AttributeRoleEnum _attributeRole;
        private bool? _inRelationship;
        private MappedProperty _mapped;

        public Property(IColumn source, Entity parent, PropertyDefinition definition)
            : base(source, definition.Name, definition.Visible, parent)
        {
            _definition = definition;

            if (this.InPrimaryKey)
                _attributeRole = AttributeRoleEnum.PrimaryKey;
            else if (this.InForeignKey)
                _attributeRole = AttributeRoleEnum.ForeignKey;
            else
                _attributeRole = AttributeRoleEnum.Field;
        }

        public new Column Source
        {
            get { return base.Source as Column; }
        }
        public override Domain Domain
        {
            get { return this.Parent.Parent; }
        }


        #region ISourceColumn Members

        public bool IsIdentity
        {
            get { return this.Source.IsIdentity; }
        }

        public bool IsNullable
        {
            get { return this.Source.IsNullable; }
        }

        public bool IsComputed
        {
            get { return this.Source.IsComputed; }
        }

        public string DotNetType
        {
            get
            {
                string nameName = this.Source.DotNetType.FullName;

                if ((this.IsNullable || this.IsIdentity) && this.Source.DotNetType.IsValueType)
                    nameName = "Nullable<" + nameName + ">";

                return nameName;
            }
        }

        public int? CharacterMaximumLength
        {
            get { return this.Source.CharacterMaximumLength; }
        }

        public int? NumericPrecision
        {
            get { return this.Source.NumericPrecision; }
        }

        public int? NumericScale
        {
            get { return this.Source.NumericScale; }
        }

        public int? DateTimePrecision
        {
            get { return this.Source.DateTimePrecision; }
        }

        public bool InUnquieIndex
        {
            get { return this.Source.InUnquieIndex; }
        }
        /// <summary>
        /// Returns true if this property has been defined as part of the entity base class otherwise returns false. The
        /// name of the property is used, not the source name for comparison. This allows source names to be mapped
        /// to a consistent base class property name.
        /// </summary>
        public bool InDefinedEntityBaseClass
        {
            get
            {
                return this.Parent.BaseEntityClassProperties.Contains(this.Name);
            }
        }

        /// <summary>
        /// Returns true if this property has been defined as part of the entity business key otherwise returns false.
        /// </summary>
        public bool InBusinessKey
        {
            get { return this.Parent.BusinessKey.Contains(this.Name); }
        }
        public bool InPrimaryKey
        {
            get { return this.Source.InPrimaryKey; }
        }
        public bool InRelationship
        {
            get
            {
                if (!_inRelationship.HasValue)
                {
                    _inRelationship = false;
                    foreach (Relationship r in this.Parent.Relationships())
                    {
                        if (r.RelationshipType == RelationshipTypeEnum.OneToOne)
                            continue;

                        foreach (Property p in r.Properties)
                            if (p.Source.Name == this.Source.Name)
                            {
                                _inRelationship = true;
                                break;
                            }
                    }
                }

                return _inRelationship.Value;
            }
        }
        public bool InRelationshipType(RelationshipTypeEnum type)
        {
            bool inRelationship = false;
            foreach (Relationship r in this.Parent.Relationships())
            {
                if (r.RelationshipType != (r.RelationshipType & type))
                    continue;

                foreach (Property p in r.Properties)
                    if (p.Source.Name == this.Source.Name)
                    {
                        inRelationship = true;
                        break;
                    }
            }
            return inRelationship;
        }
        public bool InForeignKey
        {
            get { return this.Source.InForeignKey; }
        }

        public AttributeRoleEnum AttributeRole
        {
            get { return _attributeRole; }
            internal set { _attributeRole = value; }
        }

        #endregion

        public MappedProperty Mapped
        {
            get { return _mapped; }
            set { _mapped = value; }
        }

        /// <summary>
        /// Determines if the property has a mapping to another entities property(ies)
        /// </summary>
        public bool IsMapped
        {
            get
            {
                Column sourceColumn = this.Mapped.Properties[0].Source;
                if (sourceColumn.Parent.GetType() == typeof(Table))
                    // Compare the source tables for the mapping. If they are the same then its not been mapped.
                    if (((Table)sourceColumn.Parent).Name != this.Parent.Source.Name)
                        return true;


                return this.Name != this.Mapped.Properties[0].Name;
            }
        }

        #region ICloneable Members

        public object Clone()
        {
            return new Property(this.Source, this.Parent, _definition);
        }

        #endregion

        public override string ToString()
        {
            return this.Mapped.Name;
        }
    }
}
