#region License
// (c) Garry McGlennon 2009
// 
// This project is made available under the Microsoft Public License (Ms-PL)
// Code is provided as is and with no warrenty  Use at your own risk
// 
// View the project and the latest code at http://dnpframework.codeplex.com/
// 
// All other rights reserved.
// 
// blog: www.dotNetProfessional.com/blog/
#endregion
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Reflection;
using dnp.Framework.Scripting;

namespace dnp.Framework.Collections
{
    [Serializable]
    public class MultiKeyedCollection<T> : KeyedCollection<string, T>
    {
        private List<string> _keys = new List<string>();

        /// <summary>
        /// Initialises the collection with the property names of T that form the unique key for T.
        /// </summary>
        /// <param name="key">Comma seperated list of properties that make up the key for the object T.</param>
        public MultiKeyedCollection(params string[] key)
        {
            if (key.Length == 0)
                throw new ArgumentException("Collection must have at least one property defined as its key.");

            // Verify that the keys exist!
            // We grab a copy of the properties as we will need this collection once for each key
            List<PropertyInfo> properties = new List<PropertyInfo>();

            if (typeof(T).IsInterface)
            {
                // For some reason interfaces won't report their base types so we need to treat them differently!!
                foreach (Type i in typeof(T).GetInterfaces())
                    properties.AddRange(i.GetProperties());
            }
            else
                properties.AddRange(typeof(T).GetProperties());

            foreach (string k in key)
            {
                bool keyFound = false;

                foreach (PropertyInfo p in properties)
                {
                    if (p.Name == k)
                    {
                        keyFound = true;
                        break;
                    }
                }

                if (!keyFound)
                    throw new KeyNotFoundException(
                        string.Format("Property {0} does not exist so cannot be used as a key.", k));
            }
            foreach (string k in key)
                _keys.Add(k);
        }

        protected override string GetKeyForItem(T item)
        {
            return this.GetItemSingleKey(item);
        }

        /// <summary>
        /// Returns a string list of the properties defined as the key for this collection, in the order that is required by other routines.
        /// </summary>
        public List<string> Key
        {
            get { return _keys; }
        }

        /// <summary>
        /// Returns an object T where the key matches the specified object.
        /// </summary>
        /// <param name="key">Values for the key as defined in the constructor. The order of the values must match that of the original constructor order.</param>
        /// <returns>returns object T if found otherwise throws exception</returns>
        public T this[params string[] key]
        {
            get
            {
                string singleKey = "";

                if (this.Key.Count != key.Length)
                    throw new ArgumentException(
                        string.Format(
                            "Number of keys passed don't match the key defintion. The number passsed was {0} but the number expected is {1}.",
                            key.Length, this.Key.Count), this.GetType().Name + "[]");
                try
                {
                    singleKey = this.GetItemSingleKey(key);
                    return base[singleKey];

                }
                catch (KeyNotFoundException keyEx)
                {
                    throw new KeyNotFoundException(
                        string.Format("The key {0} does not exist in the collection {1}.", singleKey, this.GetType().Name + "[]"), keyEx);
                }
            }
            set { throw new NotSupportedException(); }
        }

        /// <summary>
        /// Remove the item in the collection that matches the key
        /// </summary>
        /// <param name="key">Values for the key as defined in the constructor. The order of the values must match that of the original constructor order.</param>
        public void Remove(params string[] key)
        {
            base.Remove(this.GetItemSingleKey(key));
        }

        /// <summary>
        /// Return true if the key exits within the collection, otherwise return false.
        /// </summary>
        /// <param name="key">Values for the key as defined in the constructor. The order of the values must match that of the original constructor order.</param>
        /// <returns></returns>
        public bool Contains(params string[] key)
        {
            return base.Contains(this.GetItemSingleKey(key));
        }

        private string GetItemSingleKey(T item)
        {
            string singleKey = "";
            // We grab a copy of the properties as we will need this collection once for each key
            PropertyInfo[] properties = ((object)item).GetType().GetProperties();

            // We need to cycle through the defined keys, as the order is critical and the order of 
            // a reflected list can't be assumed.
            foreach (string key in this.Key)
                foreach (PropertyInfo p in properties)
                {
                    if (p.Name == key)
                    {
                        object value = p.GetValue(item, null);

                        if (value == null || value.ToString() == "")
                            throw new ArgumentException("The value used for the key on the entity " + p.ReflectedType.Name + " can not be null or empty.");

                        string keyValue = value.ToString();
                        singleKey += "|" + keyValue;
                        break;
                    }
                }

            return singleKey;
        }

        private string GetItemSingleKey(string[] key)
        {
            string singleKey = "";
            foreach (string k in key)
                singleKey += "|" + k;

            return singleKey;
        }

        public void AddRange(IEnumerable<T> items)
        {
            foreach (T item in items)
                this.Add(item);
        }

        public string ToString(string expression, string joiner)
        {
            return new TextList(this, expression).ToString(joiner);
        }
    }
}