#region License
// (c) Garry McGlennon 2009
// 
// This project is made available under the Microsoft Public License (Ms-PL)
// Code is provided as is and with no warrenty  Use at your own risk
// 
// View the project and the latest code at http://dnpframework.codeplex.com/
// 
// All other rights reserved.
// 
// blog: www.dotNetProfessional.com/blog/
#endregion
using System;
using System.Data;
using dnp.Framework.Configuration;
using dnp.Framework.Database.DataProvider.Configuration;
using dnp.Framework.Database.InformationSchema;
using dnp.Framework.Database.InformationSchema.Configuration;
using dnp.Framework.Reflection;

namespace dnp.Framework.Database.DatabaseExplorer
{
    public class Database : SchemaBase
    {
        private TableCollection _Tables = new TableCollection();
        private IInformationSchemaProvider _Provider;
        private DatabaseSchemaCollection _schemas = new DatabaseSchemaCollection();
        private ViewCollection _views = new ViewCollection();
        private SystemFieldCollection _systemFields = new SystemFieldCollection();
        private RoutineCollection _routines = new RoutineCollection();
        private string[] _ignoredTablesForManyToManyCalcs = new string[] { };
        #region Constructors

        public Database()
        {
        }
        public Database(string databaseName, SystemFieldCollection systemFields)
        {
            InitialiseProvider(databaseName, systemFields, null);
        }
        public Database(string databaseName, SystemFieldCollection systemFields, string[] ignoredTablesForManyToManyCalcs)
        {
            InitialiseProvider(databaseName, systemFields, ignoredTablesForManyToManyCalcs);
        }

        public Database(string providerName, string connectionString, SystemFieldCollection systemFields)
        {
            this.InitialiseProvider(providerName, connectionString, systemFields, new string[] { });
        }
        public Database(string providerName, string connectionString, SystemFieldCollection systemFields, string[] ignoredTablesForManyToManyCalcs)
        {
            this.InitialiseProvider(providerName, connectionString, systemFields, ignoredTablesForManyToManyCalcs);
        }
        public Database(IInformationSchemaProvider provider, string connectionString, SystemFieldCollection systemFields)
        {
            InitialiseProvider(provider, connectionString, systemFields, null);
        }
        public Database(IInformationSchemaProvider provider, string connectionString, SystemFieldCollection systemFields, string[] ignoredTablesForManyToManyCalcs)
        {
            InitialiseProvider(provider, connectionString, systemFields, ignoredTablesForManyToManyCalcs);
        }

        private void InitialiseProvider(string databaseName, SystemFieldCollection systemFields, string[] ignoredTablesForManyToManyCalcs)
        {
            // Assume that the naming convention for providers is the same as data providers except substitute DataProvider with InformationSchema
            DatabaseDefinitions dataProviderConfig = ConfigurationManager.RetrieveSection<DatabaseDefinitions>("dnp.Framework/DatabaseDefinitions");
            InformationSchemaProviderCollection informationSchemaConfig =
                ConfigurationManager.RetrieveSection<InformationSchemaProviderCollection>("dnp.Framework/InformationSchemaProviders");


            if (dataProviderConfig != null && informationSchemaConfig != null)
            {
                // Infer the schema provider from the data provider
                string dataProviderName = dataProviderConfig.Databases[databaseName].DataProvider;
                string schemaProviderName = informationSchemaConfig[dataProviderName].Name;
                this.InitialiseProvider(schemaProviderName, dataProviderConfig.Databases[databaseName].ConnectionString, systemFields, ignoredTablesForManyToManyCalcs);
            }
            else
                throw new ArgumentException("Unable to read Dataprovider configuration.");
        }

        private void InitialiseProvider(string providerName, string connectionString, SystemFieldCollection systemFields, string[] ignoredTablesForManyToManyCalcs)
        {
            InformationSchemaProviderCollection informationSchemaConfig =
                ConfigurationManager.RetrieveSection<InformationSchemaProviderCollection>("dnp.Framework/InformationSchemaProviders");

            string assembly = informationSchemaConfig[providerName].Assembly;
            // Based on the information gathered create an instance of the provider.
            IInformationSchemaProvider provider = null;
            provider = ((IInformationSchemaProvider)Assembly.CreateInstanceFromString(assembly));


            this.InitialiseProvider(provider, connectionString, systemFields, ignoredTablesForManyToManyCalcs);
        }

        private void InitialiseProvider(IInformationSchemaProvider provider, string connectionString, SystemFieldCollection systemFields, string[] ignoredTablesForManyToManyCalcs)
        {
            // Initialise the system fields
            if (systemFields != null)
                _systemFields = systemFields;

            if (ignoredTablesForManyToManyCalcs != null)
                _ignoredTablesForManyToManyCalcs = ignoredTablesForManyToManyCalcs;

            _Provider = provider;
            _Provider.ConnectionString = connectionString;

            this.PopulateDatabaseSchema();
        }

        #endregion

        public new Database Parent
        {
            get { return base.Parent as Database; }
        }

        #region Populate Structure

        protected void PopulateDatabaseSchema()
        {
            // Step 1: Populate Tables and their columns (basic)
            this.PopulateTablesAndColumns();

            // Step 2: Add Constraints ie Primary and Unique Keys
            this.PopulateConstraints();

            // Step 3: Add Foreign Constraints ie Primary and Unique Keys
            this.PopulateForeignKeyContraints();

            // Step 4: Populare Indexes
            this.PopulateIndexes();

            // Step 5: Populare Relationships
            this.PopulateRelationships();

            // Step 6: Populate Views & Columns
            this.PopulateViewsAndColumns();

            // Step 7: Populate Routines
            this.PopulateRoutinesAndParameters();
            this.PopulateRoutineColumns();

            // Step 8: Organise tables into their schema's
            this.PopulateSchemas();

            // Step 9: Populate Extended Properties
            this.PopulateExtendedProperties();

            // Now populate the database name
            this.Name = this.Provider.DatabaseName;
        }

        protected void PopulateTablesAndColumns()
        {
            ColumnSchemaCollection columns = this.Provider.RetrieveColumnSchema();
            ColumnSchemaCollection currentTableColumns = new ColumnSchemaCollection();

            string currentTable = "";
            string previousTable = "";

            foreach (ColumnSchema c in columns)
            {
                currentTable = c.TableSchema + ":" + c.TableName;
                if (currentTable != previousTable && previousTable != "")
                {
                    Table t = new Table(this, currentTableColumns);
                    // Define the dotNetDataType for each column in the table
                    foreach (TableColumn tc in t.Columns)
                    {
                        tc.DotNetType = this.Provider.DotNetType(tc.DataType);
                        tc.SqlType = this.Provider.SQLType(tc.ColumnSchema);
                        tc.Size = this.Provider.Size(tc.ColumnSchema);
                    }
                    this.Tables.Add(t);
                    // Now reset the collection for the next table
                    currentTableColumns = new ColumnSchemaCollection();
                }

                currentTableColumns.Add(c);

                previousTable = currentTable;
            }
            if (currentTableColumns.Count != 0)
            {
                Table t = new Table(this, currentTableColumns);
                // Define the dotNetDataType for each column in the table
                foreach (TableColumn tc in t.Columns)
                {
                    tc.SqlType = this.Provider.SQLType(tc.ColumnSchema);
                    tc.DotNetType = this.Provider.DotNetType(tc.DataType);
                    tc.Size = this.Provider.Size(tc.ColumnSchema);
                }
                this.Tables.Add(t);
            }
        }

        protected void PopulateRoutinesAndParameters()
        {
            RoutineParameterColumnSchemaCollection parameters = this.Provider.RetrieveRoutineParametersSchema();
            RoutineParameterColumnSchemaCollection currentRoutineParameters = new RoutineParameterColumnSchemaCollection();

            string currentRoutine = "";
            string previousRoutine = "";

            foreach (RoutineParameterColumnSchema p in parameters)
            {
                currentRoutine = p.TableSchema + ":" + p.TableName;
                if (currentRoutine != previousRoutine && previousRoutine != "")
                {
                    Routine r = new Routine(this, currentRoutineParameters);
                    // Define the dotNetDataType for each column in the table
                    foreach (RoutineParameterColumn rc in r.Parameters)
                    {
                        rc.DotNetType = this.Provider.DotNetType(rc.DataType);
                        rc.SqlType = this.Provider.SQLType(rc.ColumnSchema);
                        rc.Size = this.Provider.Size(rc.ColumnSchema);
                    }
                    this.Routines.Add(r);
                    // Now reset the collection for the next routine
                    currentRoutineParameters = new RoutineParameterColumnSchemaCollection();
                }

                currentRoutineParameters.Add(p);

                previousRoutine = currentRoutine;
            }
            if (currentRoutineParameters.Count != 0)
            {
                Routine r = new Routine(this, currentRoutineParameters);
                // Define the dotNetDataType for each column in the table
                foreach (RoutineParameterColumn rc in r.Parameters)
                {
                    rc.SqlType = this.Provider.SQLType(rc.ColumnSchema);
                    rc.DotNetType = this.Provider.DotNetType(rc.DataType);
                    rc.Size = this.Provider.Size(rc.ColumnSchema);
                }
                this.Routines.Add(r);
            }

        }

        protected void PopulateRoutineColumns()
        {
            ColumnSchemaCollection columns = this.Provider.RetrieveRoutineColumnSchema();
            ColumnSchemaCollection currentRoutineColumns = new ColumnSchemaCollection();

            string currentRoutine = "";
            string previousRoutine = "";

            foreach (ColumnSchema c in columns)
            {
                currentRoutine = c.TableSchema + "." + c.TableName;
                if (currentRoutine != previousRoutine && previousRoutine != "")
                {
                    Routine r;
                    if (this.Routines.Contains(previousRoutine))
                    {
                        r = this.Routines[previousRoutine];
                        r.PopulateColumns(currentRoutineColumns);
                    }
                    else
                    {
                        r = new Routine(this, currentRoutineColumns);
                        this.Routines.Add(r);
                    }
                    // Define the dotNetDataType for each column in the table
                    foreach (RoutineColumn rc in r.Columns)
                    {
                        rc.DotNetType = this.Provider.DotNetType(rc.DataType);
                        rc.SqlType = this.Provider.SQLType(rc.ColumnSchema);
                        rc.Size = this.Provider.Size(rc.ColumnSchema);
                    }
                    // Now reset the collection for the next table
                    currentRoutineColumns = new ColumnSchemaCollection();
                }

                currentRoutineColumns.Add(c);

                previousRoutine = currentRoutine;
            }
            if (currentRoutineColumns.Count != 0)
            {
                Routine r;
                if (this.Routines.Contains(currentRoutine))
                {
                    r = this.Routines[previousRoutine];
                    r.PopulateColumns(currentRoutineColumns);
                }
                else
                {
                    r = new Routine(this, currentRoutineColumns);
                    this.Routines.Add(r);
                }

                // Define the dotNetDataType for each column in the table
                foreach (RoutineColumn rc in r.Columns)
                {
                    rc.SqlType = this.Provider.SQLType(rc.ColumnSchema);
                    rc.DotNetType = this.Provider.DotNetType(rc.DataType);
                    rc.Size = this.Provider.Size(rc.ColumnSchema);
                }
            }
        }

        protected void PopulateConstraints()
        {
            ConstraintSchemaCollection constraint = this.Provider.RetrieveConstraintSchema();

            string currentConstraintName = "";
            string previousConstraintName = "";

            ConstraintSchemaCollection currentConstraint = new ConstraintSchemaCollection();

            foreach (ConstraintSchema c in constraint)
            {
                currentConstraintName = c.ConstraintName;
                if (currentConstraintName != previousConstraintName && previousConstraintName != "")
                {
                    // Locate the table that the constraint is associated with
                    if (currentConstraint.Count != 0)
                    {
                        Table table = this.Tables[currentConstraint[0].TableSchema + "." + currentConstraint[0].TableName];
                        table.Constraints.Add(new Constraint(table, currentConstraint));
                        currentConstraint = new ConstraintSchemaCollection();
                    }
                }

                currentConstraint.Add(c);

                previousConstraintName = currentConstraintName;
            }

            // The last table needs to be recorded
            if (currentConstraint.Count != 0)
            {
                // Locate the table that the constraint is associated with
                Table table = this.Tables[currentConstraint[0].TableSchema + "." + currentConstraint[0].TableName];
                table.Constraints.Add(new Constraint(table, currentConstraint));
            }
        }

        protected void PopulateIndexes()
        {
            IndexSchemaCollection Index = this.Provider.RetrieveIndexSchema();

            string currentIndexName = "";
            string previousIndexName = "";

            IndexSchemaCollection currentIndex = new IndexSchemaCollection();

            foreach (IndexSchema c in Index)
            {
                currentIndexName = c.ConstraintName;
                if (currentIndexName != previousIndexName && previousIndexName != "")
                {
                    // Locate the table that the Index is associated with
                    if (currentIndex.Count != 0)
                    {
                        string tableName = currentIndex[0].TableSchema + "." + currentIndex[0].TableName;
                        //TODO: [GM] Add suppport for Indexes on Views
                        if (this.Tables.Contains(tableName))
                        {
                            Table table = this.Tables[tableName];
                            table.Indexes.Add(new Index(table, currentIndex));
                            currentIndex = new IndexSchemaCollection();
                        }
                    }
                }

                currentIndex.Add(c);

                previousIndexName = currentIndexName;
            }

            // The last table needs to be recorded
            if (currentIndex.Count != 0)
            {
                string tableName = currentIndex[0].TableSchema + "." + currentIndex[0].TableName;
                //TODO: [GM] Add suppport for Indexes on Views
                if (this.Tables.Contains(tableName))
                {
                    // Locate the table that the Index is associated with
                    Table table = this.Tables[tableName];
                    table.Indexes.Add(new Index(table, currentIndex));
                }
            }
        }

        protected void PopulateForeignKeyContraints()
        {
            string previousConstraintName = "";
            ForeignConstraintSchemaCollection schemaConstraints = this.Provider.RetrieveForeignConstraintSchema();
            ForeignConstraintSchemaCollection currentConstraint = new ForeignConstraintSchemaCollection();

            foreach (ForeignConstraintSchema c in schemaConstraints)
            {
                if (!(c.ConstraintName == previousConstraintName || previousConstraintName == ""))
                {
                    // Locate the table that the constraint is associated with
                    if (currentConstraint.Count != 0)
                    {
                        Table table = this.Tables[currentConstraint[0].TableSchema + "." + currentConstraint[0].TableName];
                        // Record the foreign key constraint in the tables general constraints collection
                        table.Constraints.Add(new ForeignKeyConstraint(table, currentConstraint));

                        currentConstraint = new ForeignConstraintSchemaCollection();
                    }
                }

                // Record the constraint which could be part of a multi-part constraint
                currentConstraint.Add(c);

                previousConstraintName = c.ConstraintName;
            }

            // The last table needs to be recorded
            if (currentConstraint.Count != 0)
            {
                // Locate the table that the constraint is associated with
                Table table = this.Tables[currentConstraint[0].TableSchema + "." + currentConstraint[0].TableName];
                table.Constraints.Add(new ForeignKeyConstraint(table, currentConstraint));
            }
        }

        protected void PopulateRelationships()
        {
            // Convert all foreign key constraints into relationships
            foreach (Table t in this.Tables)
            {
                foreach (Constraint c in t.Constraints)
                    if (c.ConstraintType == ConstraintTypeEnum.ForeignKey)
                    {
                        ForeignKeyConstraint f = c as ForeignKeyConstraint;
                        if (f != null)
                        {
                            Relationship rel = new Relationship(t, f.PrimaryKey, c);
                            t.Relationships.Add(rel);
                        }
                    }
            }


            // Now that we have all the relationships defined, we need to check if any OnetoMany/ManyToOne relationships are really ManyToMany relationships.
            // Step 1: Cycle through the tables and identify which ones are considered Inersection tables or Simple Joins
            foreach (Table t in this.Tables)
                if (t.IsIntersectionTable)
                {
                    // Step 2: Identify the two associated tables for the Many to Many
                    Constraint c1 = t.Relationships[0].PrimaryKey;
                    Constraint c2 = t.Relationships[1].PrimaryKey;

                    //// Step 3: Re-assign the associated tables relationship foreign tables.

                    // Step 3: Re-assign the associated tables relationship foreign tables.
                    c1.Parent.Relationships[t.Relationships[0].Name].IntermediateTable = c1.Parent.Relationships[t.Relationships[0].Name].ForeignKey.Parent;
                    c1.Parent.Relationships[t.Relationships[0].Name].ForeignKey = c2;

                    c2.Parent.Relationships[t.Relationships[1].Name].IntermediateTable = c2.Parent.Relationships[t.Relationships[1].Name].ForeignKey.Parent;
                    c2.Parent.Relationships[t.Relationships[1].Name].ForeignKey = c1;
                }
        }

        protected void PopulateSchemas()
        {
            // Populate the tables
            foreach (Table t in this.Tables)
            {
                if (!this.Schemas.Contains(t.Schema))
                    this.Schemas.Add(new DatabaseSchema(t.Schema, t.Parent));

                // Add this table to the correct schema collection
                this.Schemas[t.Schema].Tables.Add(t);
            }

            // Populate Views
            foreach (View v in this.Views)
            {
                if (!this.Schemas.Contains(v.Schema))
                    this.Schemas.Add(new DatabaseSchema(v.Schema, v.Parent));

                // Add this table to the correct schema collection
                this.Schemas[v.Schema].Views.Add(v);
            }
        }

        protected void PopulateViewsAndColumns()
        {
            ColumnSchemaCollection columns = this.Provider.RetrieveViewColumnSchema();
            ColumnSchemaCollection currentViewColumns = new ColumnSchemaCollection();

            string currentView = "";
            string previousView = "";

            foreach (ColumnSchema c in columns)
            {
                currentView = c.TableSchema + ":" + c.TableName;
                if (currentView != previousView && previousView != "")
                {
                    View v = new View(this, currentViewColumns);
                    // Define the dotNetDataType for each column in the table
                    foreach (ViewColumn vc in v.Columns)
                    {
                        vc.DotNetType = this.Provider.DotNetType(vc.DataType);
                        vc.SqlType = this.Provider.SQLType(vc.ColumnSchema);
                        vc.Size = this.Provider.Size(vc.ColumnSchema);
                    }
                    this.Views.Add(v);
                    // Now reset the collection for the next table
                    currentViewColumns = new ColumnSchemaCollection();
                }

                currentViewColumns.Add(c);

                previousView = currentView;
            }
            if (currentViewColumns.Count != 0)
            {
                View v = new View(this, currentViewColumns);
                // Define the dotNetDataType for each column in the table
                foreach (ViewColumn vc in v.Columns)
                {
                    vc.SqlType = this.Provider.SQLType(vc.ColumnSchema);
                    vc.DotNetType = this.Provider.DotNetType(vc.DataType);
                    vc.Size = this.Provider.Size(vc.ColumnSchema);
                }
                this.Views.Add(v);
            }
        }

        protected void PopulateExtendedProperties()
        {
            ExtendedPropertySchemaCollection properties = this.Provider.RetrieveExtendedPropertySchema();

            foreach (ExtendedPropertySchema p in properties)
            {
                if (this.Tables.Contains(p.TableSchema + "." + p.TableName))
                {
                    Table t = this.Tables[p.TableSchema + "." + p.TableName];
                    if (string.IsNullOrEmpty(p.ColumnName))
                        t.ExtendedProperties.Add(new ExtendedProperty(p.PropertyName, p.PropertyDescription));
                    else
                        t.Columns[p.ColumnName].ExtendedProperties.Add(new ExtendedProperty(p.PropertyName, p.PropertyDescription));
                }

                if (this.Views.Contains(p.TableSchema + "." + p.TableName))
                {
                    View v = this.Views[p.TableSchema + "." + p.TableName];
                    if (string.IsNullOrEmpty(p.ColumnName))
                        v.ExtendedProperties.Add(new ExtendedProperty(p.PropertyName, p.PropertyDescription));
                    else
                        v.Columns[p.ColumnName].ExtendedProperties.Add(new ExtendedProperty(p.PropertyName, p.PropertyDescription));
                }
            }
        }

        #endregion

        #region Public Members

        public IInformationSchemaProvider Provider
        {
            get { return _Provider; }
        }

        public TableCollection Tables
        {
            get { return _Tables; }
            set { _Tables = value; }
        }

        public DatabaseSchemaCollection Schemas
        {
            get { return this._schemas; }
        }

        public ViewCollection Views
        {
            get { return _views; }
        }

        public RoutineCollection Routines
        {
            get { return _routines; }
        }

        public bool IsSystemField(string fieldName)
        {
            return this.SystemFields.Contains(fieldName);
        }

        /// <summary>
        /// Defines the system fields which are used to calcualte M:M relationships by ignoring these fields
        /// </summary>
        public SystemFieldCollection SystemFields
        {
            get { return _systemFields; }
        }

        /// <summary>
        /// Used to force a table to be ignored when calculating M:M relationships
        /// </summary>
        public string[] IgnoredTablesForManyToManyCalcs
        {
            get { return _ignoredTablesForManyToManyCalcs; }
        }

        public IDataReader ExecuteSQL(string sqlStatement)
        {
            return this.Provider.ExecuteSQL(sqlStatement);
        }

        public TableCollection OrderedTables(TableOrderEnum order)
        {
            switch (order)
            {
                case TableOrderEnum.InsertOrder:
                    return this.InsertOrderedTableList();
                case TableOrderEnum.DeleteOrder:
                    return this.DeleteOrderedTableList();
                default:
                    break;
            }

            throw new ArgumentException("Unknown type.");
        }

        private TableCollection InsertOrderedTableList()
        {
            // Make a copy of the tables we have then we'll move then out
            // as we account for them

            TableCollection unorderedTables = new TableCollection();
            TableCollection orderedTables = new TableCollection();
            TableCollection independantTables = new TableCollection();

            foreach (Table t in this.Tables)
                unorderedTables.Add(t);

            do
            {
                independantTables = GetIndependantTables(unorderedTables, orderedTables);
                // Now add the result to the ordered list and remove them from
                // the unordered list
                foreach (Table t in independantTables)
                {
                    orderedTables.Add(t);
                    unorderedTables.Remove(t);
                }
            } while (independantTables.Count != 0);

            if (unorderedTables.Count != 0)
                throw new ArgumentException(string.Format("Unable to process {0} tables. Use of this routine is therefore not recommended, please report this as a bug to the author.", unorderedTables.Count));

            return orderedTables;
        }

        private static TableCollection GetIndependantTables(TableCollection source, TableCollection exclusionList)
        {
            TableCollection independantTables = new TableCollection();
            foreach (Table t in source)
            {
                if (t.Relationships.Count > 0)
                {
                    bool isIndependant = true;
                    foreach (Relationship r in t.Relationships)
                    {
                        // Check if anything references this table
                        if (r.RelationshipType != RelationshipTypeEnum.OneToMany &&
                            r.RelationshipType != RelationshipTypeEnum.ManyToMany &&
                            r.RelationshipType != RelationshipTypeEnum.Recursive &&
                            !exclusionList.Contains(r.RightConstraint.TableFullName))
                        {
                            isIndependant = false;
                            break;
                        }
                    }

                    if (isIndependant)
                        independantTables.Add(t);
                }
                else
                    // This table has zero
                    independantTables.Add(t);
            }

            return independantTables;
        }

        private TableCollection DeleteOrderedTableList()
        {
            // This is the reverse order of the insert list
            TableCollection insertOrder = this.OrderedTables(TableOrderEnum.InsertOrder);
            TableCollection deleteOrder = new TableCollection();

            for (int i = insertOrder.Count - 1; i >= 0; i--)
            {
                deleteOrder.Add(insertOrder[i]);
            }
            return deleteOrder;
        }

        #endregion
    }
}