#region License
// (c) Garry McGlennon 2009
// 
// This project is made available under the Microsoft Public License (Ms-PL)
// Code is provided as is and with no warrenty  Use at your own risk
// 
// View the project and the latest code at http://dnpframework.codeplex.com/
// 
// All other rights reserved.
// 
// blog: www.dotNetProfessional.com/blog/
#endregion
namespace dnp.Framework.Database.DatabaseExplorer
{
    public class Relationship : SchemaBase
    {
        private Constraint _primaryKey;
        private Constraint _foreignKey;
        private RelationshipTypeEnum _relationshipType = RelationshipTypeEnum.Unknown;
        private Table _intermediateTable;
        private bool _isRecursiveMirror;

        public Relationship(Table parent, Constraint primaryKey, Constraint foreignKey)
            : base(foreignKey.Name, parent, parent.Database)
        {
            _primaryKey = primaryKey;
            _foreignKey = foreignKey;

            // Now that we have defined the relationship for this table, we need to define the opposing tables
            // relationship.
            // Don't copy the relationship if this is a OneToOne Relationship or a Recursive one
            if (this.RelationshipType != RelationshipTypeEnum.OneToOne && this.RelationshipType != RelationshipTypeEnum.Recursive)
                this.PrimaryKey.Parent.Relationships.Add(new Relationship(this.PrimaryKey.Parent, this.PrimaryKey, this.ForeignKey, false));


            if (this.RelationshipType == RelationshipTypeEnum.Recursive)
            {
                // This is a hack to change the 1:M relationship name otherwise we get a duplicate key violation. As the key is set
                // during the add to the collection and is then set we can change the name temporarily then change it back after the
                // addition otherwise the new name will flow through and still give the voilation.
                string originalName = this.ForeignKey.Name;
                this.ForeignKey.Name = originalName + "1toM";
                this.PrimaryKey.Parent.Relationships.Add(new Relationship(this.PrimaryKey.Parent, this.PrimaryKey, this.ForeignKey, true));
                this.ForeignKey.Name = originalName;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="parent"></param>
        /// <param name="primaryKey"></param>
        /// <param name="foreignKey"></param>
        /// <param name="recursive">Used to determine how to treat recursive relationships.</param>
        private Relationship(Table parent, Constraint primaryKey, Constraint foreignKey, bool recursive)
            : base(foreignKey.Name, parent, parent.Database)

        {
            // This is a duplicate as we need to create a new entity internally but don't want to create a recursive
            // loop

            _primaryKey = primaryKey;
            _foreignKey = foreignKey;

            if(recursive)
                this.IsRecursiveMirror = true;
        }

        protected bool IsRecursiveMirror
        {
            get { return _isRecursiveMirror; }
            set { _isRecursiveMirror = value; }
        }

        public new Table Parent
        {
            get{ return base.Parent as Table;}
        }
        public Constraint PrimaryKey
        {
            get { return _primaryKey; }
        }

        public Constraint ForeignKey
        {
            get { return _foreignKey; }
            internal set { _foreignKey = value; }
        }

        public Table IntermediateTable
        {
            get { return _intermediateTable; }
            internal set { _intermediateTable = value; }
        }

        public RelationshipTypeEnum RelationshipType
        {
            get
            {
                if (_relationshipType == RelationshipTypeEnum.Unknown)
                {
                    // Calculate the relationship type
                    if (this.IsManyToManyRelationship())
                        _relationshipType = RelationshipTypeEnum.ManyToMany;

                    if (_relationshipType == RelationshipTypeEnum.Unknown)
                        if (this.IsOneToOneRelationship())
                            _relationshipType = RelationshipTypeEnum.OneToOne;

                    // Recursive types are the easiest to determine as both sides have the same table (schema too!)
                    if (_relationshipType == RelationshipTypeEnum.Unknown)
                        if (this.IsRecursiveRelationship())
                            _relationshipType = RelationshipTypeEnum.Recursive;

                    if (_relationshipType == RelationshipTypeEnum.Unknown)
                        if (this.IsManyToOneRelationship())
                            _relationshipType = RelationshipTypeEnum.ManyToOne;

                    if (_relationshipType == RelationshipTypeEnum.Unknown)
                        if (this.IsOneToManyRelationship())
                            _relationshipType = RelationshipTypeEnum.OneToMany;
                }
                return this._relationshipType;
            }
        }

        public Constraint LeftConstraint
        {
            get
            {
                if(this.RelationshipType == RelationshipTypeEnum.OneToMany || this.RelationshipType == RelationshipTypeEnum.ManyToMany)
                    return this.PrimaryKey;
                else
                    return this.ForeignKey;
            }
        }

        public Constraint RightConstraint
        {
            get
            {
                if (this.RelationshipType == RelationshipTypeEnum.OneToMany || this.RelationshipType == RelationshipTypeEnum.ManyToMany)
                    return this.ForeignKey;
                else
                    return this.PrimaryKey;
            }
        }

        /// <summary>
        /// Determines if the current relationship is a Recursive one. This is where both sides of the relationship use the same table.
        /// </summary>
        /// <returns></returns>
        private bool IsRecursiveRelationship()
        {
            return !this.IsRecursiveMirror && this.ForeignKey.Parent.FullName == this.PrimaryKey.Parent.FullName;
        }

        private bool IsManyToOneRelationship()
        {
            return !this.IsRecursiveMirror && this.ForeignKey.Parent.FullName == this.Parent.FullName;
        }

        private bool IsOneToManyRelationship()
        {
            return this.IsRecursiveMirror || this.PrimaryKey.Parent.FullName == this.Parent.FullName;
        }

        private bool IsManyToManyRelationship()
        {
            return this.IntermediateTable != null;
        }
        private bool IsOneToOneRelationship()
        {
            // First check that the number of keys match between Primary and Foreign
            // It is possible for the foreign tables parent not to have a primary key!
            if (this.ForeignKey.Parent.HasPrimaryKey && this.PrimaryKey.Columns.Count == this.ForeignKey.Parent.PrimaryKey.Columns.Count)
            {
                foreach (TableColumn c in this.ForeignKey.Columns)
                {
                    if (!c.InPrimaryKey)
                        return false;
                }

                return true;
            }
            else
                return false;
        }
    }
}