#region License
// (c) Garry McGlennon 2009
// 
// This project is made available under the Microsoft Public License (Ms-PL)
// Code is provided as is and with no warrenty  Use at your own risk
// 
// View the project and the latest code at http://dnpframework.codeplex.com/
// 
// All other rights reserved.
// 
// blog: www.dotNetProfessional.com/blog/
#endregion
using System;
using System.Collections;
using dnp.Framework.Database.DatabaseExplorer;
using dnp.DomainModeler;
using dnp.ServiceModel.Configuration;
using dnp.Framework.Scripting;
using Relationship = dnp.DomainModeler.Relationship;

namespace dnp.ServiceModel
{
    public class BusinessService
    {
        private string _name = "";
        private ServiceMethodCollection _methods = new ServiceMethodCollection();
        private BusinessServiceDefinition _definition;
        private BusinessLayer _parent;

        public BusinessService(BusinessLayer parent, BusinessServiceDefinition definition)
        {
            _name = definition.Name;
            _definition = definition;
            _parent = parent;
            this.Initialize();
        }

        private void Initialize()
        {
            this.CreateServiceMethods();
            this.CreateExposedEntities();
        }

        private void CreateServiceMethods()
        {
            // Create the methods that have been defined by the user
            foreach (ServiceMethodDefinition methodDefiniton in _definition.ServiceMethods)
            {
                Entity sourceEntity = null;
                string storedProcedureName = methodDefiniton.StoredProcedureName;
                string name = methodDefiniton.Name;
                ServiceReturnType returnType = new ServiceReturnType(this.Model, methodDefiniton.ReturnType.IsEnumerable,
                    methodDefiniton.ReturnType.IsKnownType, methodDefiniton.ReturnType.Namespace, methodDefiniton.ReturnType.Type);
                ServiceFeatureEnum serviceFeature = methodDefiniton.ServiceFeature;
                if (!string.IsNullOrEmpty(methodDefiniton.SourceEntityName))
                    sourceEntity = this.Parent.Domain.Entities[methodDefiniton.SourceEntityName];

                ServiceMethod method = new ServiceMethod(this, storedProcedureName, serviceFeature, name, methodDefiniton.ServiceMethodType, true, returnType, sourceEntity);
                // Now add the parameters
                foreach (ServiceMethodParameterDefinition p in methodDefiniton.Parameters)
                    method.Parameters.Add(new ServiceMethodParameter(p.Name, p.RelativeNamespace, p.DotNetType));

                // Now add paramter mappings
                foreach (ParameterMapDefinition p in methodDefiniton.StoredProcedureMapping)
                    method.StoredProcedureMapping.Add(new ParameterMap(p.ParameterName, p.PropertyMapping, Type.GetType(p.DotNetType.Contains(".") ? p.DotNetType : "System." + p.DotNetType), false, p.IsNullable, p.NullablePropertyParameterCheck, p.DbTypeSize));

                // Now add any return type property mappings
                foreach (ServiceReturnTypePropertyMapDefinition p in methodDefiniton.ReturnTypeParamterMapping)
                    method.ReturnTypePropertyMapping.Add(new ServiceReturnTypePropertyMap(p.DotNetType, p.IsIntrinsicType, p.FieldName, p.IsNullable, p.PropertyMapping));

                this.Methods.Add(method);
            }
        }

        #region Define Exposed Entities

        private void CreateExposedEntities()
        {
            // Create the methods that have been defined by the user
            foreach (ExposedEntityDefinition methodDefiniton in _definition.ExposedEntities)
            {
                Entity sourceEntity = this.Parent.Domain.Entities.FindBySourceObjectName(methodDefiniton.SourceObjectName);
                // As Enums are immutable (ie can't be changed) we won't create any ability to save/edit them.
                //if (!sourceEntity.Visible || sourceEntity.EntityType == EntityTypeEnum.Enum)
                if (sourceEntity == null || sourceEntity.EntityType == EntityTypeEnum.Enum)
                    continue;

                // The following types only apply to entities if thier source is a Table
                if (sourceEntity.Source.GetType() == typeof(TableSource))
                {
                    if (methodDefiniton.IncludeStore && sourceEntity.HasPrimaryKey)
                    {
                        this.CreateExposedEntity_StoreSingle(methodDefiniton);
                        this.CreateExposedEntity_StoreMultiple(methodDefiniton);
                        this.CreateExposedEntity_StoreManyToMany(methodDefiniton);
                        this.CreateExposedEntity_StoreMultipleManyToMany(methodDefiniton);
                    }

                    if (methodDefiniton.IncludeRetrieveMultiple)
                        this.CreateExposedEntity_SelectMultiple(methodDefiniton);

                    if (methodDefiniton.IncludeExists)
                        this.CreateExposedEntity_SelectExists(methodDefiniton);

                    if (methodDefiniton.IncludeRetrieveSingle)
                        this.CreateExposedEntity_SelectSingle(methodDefiniton);

                    if (methodDefiniton.IncludeRetrieveRelationship)
                        this.CreateExposedEntity_SelectRelationship(methodDefiniton);
                }
                else
                {
                    if (methodDefiniton.IncludeRetrieveMultiple)
                        this.CreateExposedEntity_SelectMultipleRoutine(methodDefiniton);
                }
            }
        }

        internal void ProcessRelationships()
        {
            // Each entity has a number of different method types the following processes each type in turn.
            // should a service not support a given type it is simply ignored.
            foreach (ServiceMethod sm in this.Methods)
            {
                ExposedEntity_StoreSingle_AddRelatedMethods(sm);
                ExposedEntity_StoreMultiple_AddRelatedMethods(sm);
                ExposedEntity_StoreManyToMany_AddRelatedMethods(sm);
            }

            // Many to Many relationships require the relationship detail to 
            // define their child relationships too. This loop reprocesses all
            // exposed entities that have M:M relationships.
            foreach (ExposedEntityDefinition methodDefiniton in _definition.ExposedEntities)
            {
                Entity source = this.Parent.Domain.Entities.FindBySourceObjectName(methodDefiniton.SourceObjectName);
                if (source == null || source.EntityType == EntityTypeEnum.Enum)
                    continue;

                foreach (Relationship r in source.Relationships(RelationshipTypeEnum.ManyToMany))
                {
                    if (r.RelatedEntity.EntityType == EntityTypeEnum.Enum)
                        continue;

                    ExposedEntity_StoreMultipleManyToMany_AddRelatedMethods(r);
                }
            }
        }

        # region .   ExposedEntity - Store.Single

        private void CreateExposedEntity_StoreSingle(ExposedEntityDefinition exposedEntityDefinition)
        {
            // Step 1: Create the service method
            ServiceMethod serviceMethod = this.ExposedEntity_CreateServiceMethod(ServiceFeatureEnum.StoreSingle, ServiceMethodTypeEnum.Store, exposedEntityDefinition, false);

            // Step 2: Create Service Method Parameters
            serviceMethod.Parameters.Add(new ServiceMethodParameter("entity", serviceMethod.ReturnType.Namespace, serviceMethod.ReturnType.Type));

            // Step 3: Map Service method Parameters to Stored Procedure parameters
            Entity source = serviceMethod.ReturnType.KnownType;
            this.ExposedEntity_MapStoredProcedureToEntity(serviceMethod, source);

            // Step 4: Add Related methods (this is handled by the ExposedEntity_ProcessRelatedMethods method
            //         This must be done after all service methods have been defined as we map
            //         back to these methods.

            // Add the service method to the service definition
            this.Methods.Add(serviceMethod);

        }
        private void ExposedEntity_StoreSingle_AddRelatedMethods(ServiceMethod serviceMethod)
        {
            if (serviceMethod.ServiceFeature == ServiceFeatureEnum.StoreSingle)
            {
                // Add any derived type relationships.
                foreach (Entity derivedEntity in serviceMethod.ReturnType.KnownType.DerivedClasses)
                {
                    ServiceMethodRelationship relatedMethod;
                    relatedMethod = this.ExposedEntity_AddRelatedMethod(derivedEntity, ServiceFeatureEnum.StoreSingle, RelationshipTypeEnum.Derived, serviceMethod);
                    ExposedEntity_AddRelatedMethodParameterMapping(relatedMethod, DomainModeler.Model.FixNamespace("dl." + derivedEntity.RelativeNamespace), derivedEntity.Name, "entity", serviceMethod);

                }

                foreach (Relationship r in serviceMethod.ReturnType.KnownType.Relationships())
                {
                    if (r.RelatedEntity.EntityType == EntityTypeEnum.Enum)
                        continue;

                    string parameterNameMapping = "";
                    string dotNetTypeNamespace = DomainModeler.Model.FixNamespace("dl." + r.RelatedEntity.RelativeNamespace);
                    string dotNetType = r.RelatedEntity.Name;

                    ServiceFeatureEnum serviceMethodFeature = ServiceFeatureEnum.Unknown;
                    switch (r.RelationshipType)
                    {
                        case RelationshipTypeEnum.OneToOne:
                            parameterNameMapping = "entity";
                            serviceMethodFeature = ServiceFeatureEnum.StoreSingle;
                            break;
                        case RelationshipTypeEnum.Recursive:
                        case RelationshipTypeEnum.ManyToOne:
                            // We only support M:1 where its a single key or not an Entity
                            if (r.Properties.Count > 1 || r.RelatedEntity.EntityType != EntityTypeEnum.Entity)
                                continue;

                            parameterNameMapping = "entity." + r.Name;
                            serviceMethodFeature = ServiceFeatureEnum.StoreSingle;
                            break;
                        case RelationshipTypeEnum.OneToMany:
                            parameterNameMapping = "entity." + r.Name;
                            serviceMethodFeature = ServiceFeatureEnum.StoreMultiple;
                            break;
                        case RelationshipTypeEnum.ManyToMany:
                            parameterNameMapping = "entity." + r.Name;
                            serviceMethodFeature = ServiceFeatureEnum.StoreManyToManyMultiple;
                            break;
                        default:
                            continue;
                    }

                    ServiceMethodRelationship relatedMethod;
                    relatedMethod = this.ExposedEntity_AddRelatedMethod(r, serviceMethodFeature, serviceMethod);
                    if (serviceMethodFeature == ServiceFeatureEnum.StoreManyToManyMultiple)
                        relatedMethod.ParameterMapping.Add(new ServiceMethodParameter("entity", "dl", r.Parent.RootEntity.RelativeName + "BusinessKey"));

                    if (r.RelationshipType == RelationshipTypeEnum.OneToMany || r.RelationshipType == RelationshipTypeEnum.ManyToMany)
                        dotNetType += "Collection";

                    ExposedEntity_AddRelatedMethodParameterMapping(relatedMethod, dotNetTypeNamespace, dotNetType, parameterNameMapping, serviceMethod);
                }
            }
        }

        #endregion

        #region .   ExposedEntity - Store.Multiple

        private void CreateExposedEntity_StoreMultiple(ExposedEntityDefinition exposedEntityDefinition)
        {
            // Step 1: Create the service method
            ServiceMethod serviceMethod = this.ExposedEntity_CreateServiceMethod(ServiceFeatureEnum.StoreMultiple, ServiceMethodTypeEnum.Store, exposedEntityDefinition, true);

            // Step 2: Create Service Method Parameters
            serviceMethod.Parameters.Add(new ServiceMethodParameter("entities", serviceMethod.ReturnType.Namespace, serviceMethod.ReturnType.Type + "Collection"));

            // Step 3: Map Service method Parameters to Stored Procedure parameters
            Entity source = serviceMethod.ReturnType.KnownType;
            this.ExposedEntity_MapStoredProcedureToEntity(serviceMethod, source);

            // Step 4: Add Related methods (this is handled by the ExposedEntity_ProcessRelatedMethods method
            //         This must be done after all service methods have been defined as we map
            //         back to these methods.

            // Add the service method to the service definition
            this.Methods.Add(serviceMethod);

        }
        private void ExposedEntity_StoreMultiple_AddRelatedMethods(ServiceMethod serviceMethod)
        {
            if (serviceMethod.ServiceFeature == ServiceFeatureEnum.StoreMultiple)
            {
                // Add Single-store as a relationship for store multiple
                ServiceMethodRelationship relatedMethod;
                relatedMethod = this.ExposedEntity_AddRelatedMethod(serviceMethod.ReturnType.KnownType, ServiceFeatureEnum.StoreSingle, RelationshipTypeEnum.ManyToOne, serviceMethod);
                ExposedEntity_AddRelatedMethodParameterMapping(relatedMethod, DomainModeler.Model.FixNamespace("dl." + serviceMethod.ReturnType.KnownType.RelativeNamespace), serviceMethod.ReturnType.KnownType.Name, "entity", serviceMethod);
            }
        }

        #endregion

        #region .   ExposedEntity - Store.Multiple.ManyToMany

        private void CreateExposedEntity_StoreMultipleManyToMany(ExposedEntityDefinition exposedEntityDefinition)
        {

            Entity source = this.Parent.Domain.Entities.FindBySourceObjectName(exposedEntityDefinition.SourceObjectName);

            foreach (Relationship r in source.Relationships(RelationshipTypeEnum.ManyToMany))
            {
                if (r.RelatedEntity.EntityType == EntityTypeEnum.Enum)
                    continue;

                // Step 1: Create the service method
                // This creates a single store for a M:M the collection version is created elewhere?
                string storedProcedureName = this.Parent.MetaData.ServiceMethods(r, ServiceFeatureEnum.StoreManyToManyMultiple).StoredProcedureName;
                string methodName = this.Parent.MetaData.ServiceMethods(r, ServiceFeatureEnum.StoreManyToManyMultiple).MethodName;
                ServiceReturnType returnType = new ServiceReturnType(this.Parent.Model, false, true, DomainModeler.Model.FixNamespace("dl." + r.RelatedEntity.RelativeNamespace), r.RelatedEntity.Name);
                ServiceMethod serviceMethod = new ServiceMethod(this, storedProcedureName, ServiceFeatureEnum.StoreManyToManyMultiple, methodName, ServiceMethodTypeEnum.Store, false, returnType, returnType.KnownType);

                // Step 2: Create Service Method Parameters
                serviceMethod.Parameters.Add(new ServiceMethodParameter("parent", DomainModeler.Model.FixNamespace("dl." + r.Parent.RootEntity.RelativeNamespace), r.Parent.RootEntity.Name + "BusinessKey"));
                serviceMethod.Parameters.Add(new ServiceMethodParameter("entities", DomainModeler.Model.FixNamespace("dl." + r.RelatedEntity.RelativeNamespace), r.RelatedEntity.Name + "Collection"));

                // Step 3: Map Service method Parameters to Stored Procedure parameters
                this.ExposedEntity_MapStoredProcedureToEntity(serviceMethod, r.Parent.RootEntity.BusinessKey, "Parent", "parent");
                this.ExposedEntity_MapStoredProcedureToEntity(serviceMethod, r.RelatedEntity);

                // Step 4: Add Related methods (this is handled by the ExposedEntity_ProcessRelatedMethods method
                //         This must be done after all service methods have been defined as we map
                //         back to these methods.

                // Add the service method to the service definition
                this.Methods.Add(serviceMethod);
            }
        }
        private void ExposedEntity_StoreMultipleManyToMany_AddRelatedMethods(Relationship r)
        {
            // Locate the service method that we want to add related methods to
            string serviceMethodName = this.Parent.MetaData.ServiceMethods(r, ServiceFeatureEnum.StoreManyToManyMultiple).MethodName;
            ServiceMethod serviceMethod = this.Methods[serviceMethodName];

            // Add ManyToManySingle-store as a relationship for store multiple
            ServiceMethodRelationship relatedMethod;
            relatedMethod = this.ExposedEntity_AddRelatedMethod(r, ServiceFeatureEnum.StoreManyToMany, RelationshipTypeEnum.ManyToOne, serviceMethod);
            ExposedEntity_AddRelatedMethodParameterMapping(relatedMethod, DomainModeler.Model.FixNamespace("dl." + serviceMethod.ReturnType.KnownType.RelativeNamespace), serviceMethod.ReturnType.KnownType.RootEntity.Name + "BusinessKey", "entity", serviceMethod);
        }

        #endregion

        #region .   ExposedEntity - Store.ManyToMany

        private void CreateExposedEntity_StoreManyToMany(ExposedEntityDefinition exposedEntityDefinition)
        {

            Entity source = this.Parent.Domain.Entities.FindBySourceObjectName(exposedEntityDefinition.SourceObjectName);

            foreach (Relationship r in source.Relationships(RelationshipTypeEnum.ManyToMany))
            {
                if(r.RelatedEntity == null)
                    throw new ArgumentException(string.Format("The Entity {0} is marked as having a M:M relationship, however it has no related entities. This maybe due to the join table not really being a join table but an entity table. Check that the intermdediate table is a true join table.", source.Name));

                if (r.RelatedEntity.EntityType == EntityTypeEnum.Enum)
                    continue;

                // Step 1: Create the service method
                // This creates a single store for a M:M the collection version is created elewhere?
                string storedProcedureName = this.Parent.MetaData.ServiceMethods(r, ServiceFeatureEnum.StoreManyToMany).StoredProcedureName;
                string methodName = this.Parent.MetaData.ServiceMethods(r, ServiceFeatureEnum.StoreManyToMany).MethodName;
                ServiceReturnType returnType = new ServiceReturnType(this.Parent.Model, false, true, DomainModeler.Model.FixNamespace("dl." + r.RelatedEntity.RelativeNamespace), r.RelatedEntity.Name);
                ServiceMethod serviceMethod = new ServiceMethod(this, storedProcedureName, ServiceFeatureEnum.StoreManyToMany, methodName, ServiceMethodTypeEnum.Store, false, returnType, returnType.KnownType);

                // Step 2: Create Service Method Parameters
                serviceMethod.Parameters.Add(new ServiceMethodParameter("parent", DomainModeler.Model.FixNamespace("dl." + r.Parent.RootEntity.RelativeNamespace), r.Parent.RootEntity.Name + "BusinessKey"));
                serviceMethod.Parameters.Add(new ServiceMethodParameter("entity", DomainModeler.Model.FixNamespace("dl." + r.RelatedEntity.RelativeNamespace), r.RelatedEntity.RootEntity.Name + "BusinessKey"));

                // Step 3: Map Service method Parameters to Stored Procedure parameters
                this.ExposedEntity_MapStoredProcedureToEntity(serviceMethod, r.Parent.RootEntity.BusinessKey, "Parent_", "parent");
                this.ExposedEntity_MapStoredProcedureToEntity(serviceMethod, r.RelatedEntity.BusinessKey, "Child_", "entity");

                // Due to not having the intermediate table as an entity, the generic MapStoredProcedure routine will incorrectly define keys that are identities as output
                // Here we convert them all to no being output as an intermediate table can never have an identity definition.
                foreach (ParameterMap p in serviceMethod.StoredProcedureMapping)
                    p.IsOutput = false;

                // Step 4: Add Related methods (this is handled by the ExposedEntity_ProcessRelatedMethods method
                //         This must be done after all service methods have been defined as we map
                //         back to these methods.

                // Add the service method to the service definition
                this.Methods.Add(serviceMethod);
            }
        }
        private void ExposedEntity_StoreManyToMany_AddRelatedMethods(ServiceMethod serviceMethod)
        {
            if (serviceMethod.ServiceFeature == ServiceFeatureEnum.StoreManyToMany)
            {
                // Add M:M
                foreach (Relationship r in serviceMethod.ReturnType.KnownType.Relationships(RelationshipTypeEnum.ManyToMany))
                {
                    if (r.RelatedEntity.EntityType == EntityTypeEnum.Enum)
                        continue;

                    // NOTE: The return type is the related type as such for naming and service method definitions we have to look backwards
                    //       that is our 'related' entity is the known type and the 'parent' is the related entity based on the relationship instance.
                    ServiceMethodRelationship relatedMethod;
                    relatedMethod = this.ExposedEntity_AddRelatedMethod(r.Parent, ServiceFeatureEnum.StoreSingle, r.RelationshipType, serviceMethod);
                    ExposedEntity_AddRelatedMethodParameterMapping(relatedMethod, DomainModeler.Model.FixNamespace("dl." + r.Parent.RelativeNamespace), r.Parent.Name, "entity", serviceMethod);
                }
            }
        }

        #endregion

        #region .   ExposedEntity - Select.Multiple

        private void CreateExposedEntity_SelectMultiple(ExposedEntityDefinition exposedEntityDefinition)
        {
            // Step 1: Create the service method
            ServiceMethod serviceMethod = this.ExposedEntity_CreateServiceMethod(ServiceFeatureEnum.RetrieveMultiple, ServiceMethodTypeEnum.RetrieveSearchResult, exposedEntityDefinition, true);

            // Step 2: Create Service Method Parameters
            serviceMethod.Parameters.Add(new ServiceMethodParameter("searchCriteria", "", "sl.SearchCriteria"));

            // Step 3: Map Service method Parameters to Stored Procedure parameters

            // This is a specialized procedure which has no input that is specific to a given entity
            // as such this step is being omitted

            // Step 4: Add Related methods (this is handled by the ExposedEntity_ProcessRelatedMethods method
            //         This must be done after all service methods have been defined as we map
            //         back to these methods.

            // This step is not required as this method doesn't have any related methods.

            // Add the service method to the service definition
            this.Methods.Add(serviceMethod);
        }

        #endregion

        #region .   ExposedEntity - Select.Relationship

        private void CreateExposedEntity_SelectRelationship(ExposedEntityDefinition exposedEntityDefinition)
        {
            Entity source = this.Parent.Domain.Entities.FindBySourceObjectName(exposedEntityDefinition.SourceObjectName);

            foreach (Relationship r in source.Relationships(RelationshipTypeEnum.ManyToMany | RelationshipTypeEnum.OneToMany))
            {
                if (r.RelatedEntity.EntityType == EntityTypeEnum.Enum)
                    continue;

                // Step 1: Create the service method
                string storedProcedureName = this.Parent.MetaData.ServiceMethods(r, ServiceFeatureEnum.RetrieveRelationship).StoredProcedureName;
                string methodName = this.Parent.MetaData.ServiceMethods(r, ServiceFeatureEnum.RetrieveRelationship).MethodName;
                ServiceReturnType returnType = new ServiceReturnType(this.Parent.Model, true, true, DomainModeler.Model.FixNamespace("dl." + r.RelatedEntity.RelativeNamespace), r.RelatedEntity.Name);
                ServiceMethod serviceMethod = new ServiceMethod(this, storedProcedureName, ServiceFeatureEnum.RetrieveRelationship, methodName, ServiceMethodTypeEnum.RetrieveSearchRelationship, false, returnType, r.Parent);

                // Step 2: Create Service Method Parameters
                serviceMethod.Parameters.Add(new ServiceMethodParameter("relationshipSearch", "", string.Format("sl.SearchRelationship<{0}BusinessKey>", DomainModeler.Model.FixNamespace("dl." + r.Parent.RootEntity.RelativeName))));

                // Step 3: Map Service method Parameters to Stored Procedure parameters
                this.ExposedEntity_MapStoredProcedureToEntity(serviceMethod, source.BusinessKey, "", "businessKey");

                // Step 4: Add Related methods (this is handled by the ExposedEntity_ProcessRelatedMethods method
                //         This must be done after all service methods have been defined as we map
                //         back to these methods.

                // Add the service method to the service definition
                this.Methods.Add(serviceMethod);
            }
        }

        #endregion

        #region .   ExposedEntity - Select.Exists

        private void CreateExposedEntity_SelectExists(ExposedEntityDefinition exposedEntityDefinition)
        {
            // Step 1: Create the service method
            Entity source = this.Parent.Domain.Entities.FindBySourceObjectName(exposedEntityDefinition.SourceObjectName);
            string storedProcedureName = this.Parent.MetaData.ServiceMethods(source, ServiceFeatureEnum.Exists).StoredProcedureName;
            string methodName = this.Parent.MetaData.ServiceMethods(source, ServiceFeatureEnum.Exists).MethodName;
            ServiceReturnType returnType = new ServiceReturnType(this.Parent.Model, false, false, "", typeof(bool).Name);
            ServiceMethod serviceMethod = new ServiceMethod(this, storedProcedureName, ServiceFeatureEnum.Exists, methodName, ServiceMethodTypeEnum.RetrieveScalar, false, returnType, source);

            // Step 2: Create Service Method Parameters
            serviceMethod.Parameters.Add(new ServiceMethodParameter("businessKey", DomainModeler.Model.FixNamespace("dl." + source.RootEntity.RelativeNamespace), source.RootEntity.Name + "BusinessKey"));


            // Step 3: Map Service method Parameters to Stored Procedure parameters
            this.ExposedEntity_MapStoredProcedureToEntity(serviceMethod, source.BusinessKey, "", "businessKey");

            // Step 4: Add Related methods (this is handled by the ExposedEntity_ProcessRelatedMethods method
            //         This must be done after all service methods have been defined as we map
            //         back to these methods.

            // This step is not required as this method doesn't have any related methods.

            // Add the service method to the service definition
            this.Methods.Add(serviceMethod);
        }

        #endregion

        #region .   ExposedEntity - Select.Single

        private void CreateExposedEntity_SelectSingle(ExposedEntityDefinition exposedEntityDefinition)
        {
            // Step 1: Create the service method
            ServiceMethod serviceMethod = this.ExposedEntity_CreateServiceMethod(ServiceFeatureEnum.RetrieveSingle, ServiceMethodTypeEnum.Retrieve, exposedEntityDefinition, false);

            // Step 2: Create Service Method Parameters
            serviceMethod.Parameters.Add(new ServiceMethodParameter("businessKey", DomainModeler.Model.FixNamespace("dl." + serviceMethod.SourceEntity.RootEntity.RelativeNamespace), serviceMethod.SourceEntity.RootEntity.Name + "BusinessKey"));

            // Step 3: Map Service method Parameters to Stored Procedure parameters
            this.ExposedEntity_MapStoredProcedureToEntity(serviceMethod, serviceMethod.SourceEntity.BusinessKey, "", "businessKey");

            // Step 4: Define the mapping between the returned data and the entity (for population)
            // TODO: [GM] This can only be used for the query implementation and not the code implementation, due to the lack of derived type support.
            this.MapStoredProcedureResultToEntity(serviceMethod);

            // Step 5: Add Related methods (this is handled by the ExposedEntity_ProcessRelatedMethods method
            //         This must be done after all service methods have been defined as we map
            //         back to these methods.

            // This step is not required as this method doesn't have any related methods.

            // Add the service method to the service definition
            this.Methods.Add(serviceMethod);
        }

        private void MapStoredProcedureResultToEntity(ServiceMethod serviceMethod)
        {
            foreach (Property p in serviceMethod.SourceEntity.Properties(PropertyViewEnum.Entity, PropertyDepthFilterEnum.All))
            {
                // The name of the property and the name of the field used is not a simple p.Name due to mapping issues.
                string propertyName = "";
                string fieldName = "";
                string dataType = "";
                bool isIntrinsicType = true;

                string mappedName = p.Mapped.Name;
                if (serviceMethod.SourceEntity.IsDerivedClassProperty(p))
                    mappedName = p.Parent.Name + mappedName;

                if (!p.IsMapped)
                {
                    propertyName = p.Mapped.Name;
                    fieldName = mappedName;
                    dataType = p.Source.DotNetType.ToString();
                }
                else
                {
                    foreach (Property mp in p.Mapped.Properties)
                    {
                        if (mp.IsMapped)
                        {
                            if (!mp.InBusinessKey && mp.Mapped.Properties[0].Mapped.RelatedPrimaryKey.Parent.EntityType != EntityTypeEnum.Entity)
                                propertyName = p.Mapped.Name;
                            else
                                propertyName = p.Mapped.Name + "." + mp.Mapped.Properties[0].Mapped.Name;

                            fieldName = mappedName + "_" + mp.Mapped.Properties[0].Mapped.Name;
                            if (mp.Mapped.Properties[0].Mapped.RelatedPrimaryKey.Parent.EntityType == EntityTypeEnum.Enum)
                            {
                                dataType = "dl." + mp.Mapped.Properties[0].Mapped.RelatedPrimaryKey.Parent.RelativeName;
                                isIntrinsicType = false;
                            }
                            else
                                dataType = mp.Mapped.Properties[0].Source.DotNetType.ToString();
                        }
                        else
                        {
                            if (mp.Parent.EntityType != EntityTypeEnum.Entity)
                                propertyName = p.Mapped.Name;
                            else
                                propertyName = p.Mapped.Name + "." + mp.Mapped.Name;

                            fieldName = mappedName + "_" + mp.Mapped.Name;

                            if (mp.Parent.EntityType == EntityTypeEnum.Enum)
                            {
                                dataType = "dl." + mp.Parent.RelativeName;
                                isIntrinsicType = false;
                            }
                            else
                                dataType = mp.Source.DotNetType.ToString();

                        }
                    }
                }

                serviceMethod.ReturnTypePropertyMapping.Add(new ServiceReturnTypePropertyMap(dataType, isIntrinsicType, fieldName, p.IsNullable, propertyName));
            }
        }

        #endregion

        #region .   ExposedEntity - Common
        private ServiceMethod ExposedEntity_CreateServiceMethod(ServiceFeatureEnum serviceMethodFeature, ServiceMethodTypeEnum serviceMethodType, ExposedEntityDefinition exposedEntityDefinition, bool isEnumerable)
        {
            Entity source = this.Parent.Domain.Entities.FindBySourceObjectName(exposedEntityDefinition.SourceObjectName);
            string storedProcedureName = this.Parent.MetaData.ServiceMethods(source, serviceMethodFeature).StoredProcedureName;
            string methodName = this.Parent.MetaData.ServiceMethods(source, serviceMethodFeature).MethodName;
            ServiceReturnType returnType = new ServiceReturnType(this.Parent.Model, isEnumerable, true, DomainModeler.Model.FixNamespace("dl." + source.RelativeNamespace), source.Name);

            return new ServiceMethod(this, storedProcedureName, serviceMethodFeature, methodName, serviceMethodType, false, returnType, source);
        }

        /// <summary>
        /// Given an entity generate mappings to the auto-generated stored procedure.
        /// </summary>
        /// <param name="serviceMethod">service method to apply mappings</param>
        /// <param name="source">the entity that is the source of the mappings</param>
        private void ExposedEntity_MapStoredProcedureToEntity(ServiceMethod serviceMethod, Entity source)
        {
            // Get a collection of properties that match those of the stored procedure
            ViewPropertyCollection properties = source.Properties(PropertyViewEnum.Entity, PropertyDepthFilterEnum.IncludeBaseClass);

            ExposedEntity_MapStoredProcedureToEntity(serviceMethod, properties, "", "entity");
        }

        private void ExposedEntity_MapStoredProcedureToEntity(ServiceMethod serviceMethod, IEnumerable properties, string parameterPrefix, string propertyMappingPrefix)
        {
            if (!propertyMappingPrefix.EndsWith("."))
                propertyMappingPrefix += ".";

            ParameterMap mappedParameter;
            // We need to process each property of the entity.
            foreach (Property p in properties)
            {
                // Some fields we ignore as they have no storage ability
                if (p.IsComputed)
                    continue;


                if (p.IsMapped && p.Mapped.RelatedPrimaryKey.Parent.EntityType == EntityTypeEnum.Entity)
                    foreach (Property businessKey in p.Mapped.Properties)
                    {

                        // As some properties can be mapped more than once we need to determine which is the real properties to use.
                        if (p.Mapped.Properties[0].IsMapped && p.Mapped.Properties[0].Mapped.RelatedPrimaryKey.Parent.EntityType != EntityTypeEnum.Entity)
                            mappedParameter = new ParameterMap(parameterPrefix + p.Mapped.Name + businessKey.Mapped.Name, propertyMappingPrefix + p.Mapped.Name + "." + businessKey.Mapped.Name,
                                                               businessKey.Mapped.Properties[0].Source.DotNetType, IsOutputField(p), (p.IsNullable && p.Mapped.Properties[0].DotNetType != "System.String" || p.IsMapped), "entity." + p.Mapped.Name, p.Mapped.Properties[0].Source.Size);
                        else
                            mappedParameter = new ParameterMap(parameterPrefix + p.Mapped.Name + businessKey.Name, propertyMappingPrefix + p.Mapped.Name + "." + businessKey.Name,
                                                               businessKey.Source.DotNetType, IsOutputField(p), (p.IsNullable && p.DotNetType != "System.String" || p.IsMapped), "entity." + p.Mapped.Name, p.Source.Size);
                        serviceMethod.StoredProcedureMapping.Add(mappedParameter);
                    }
                else
                {
                    if (p.IsMapped)
                        mappedParameter = new ParameterMap(parameterPrefix + p.Mapped.Name + p.Mapped.Properties[0].Name, propertyMappingPrefix + p.Mapped.Name, p.Mapped.Properties[0].Source.DotNetType,
                                   IsOutputField(p), (p.IsNullable && p.DotNetType != "System.String"), propertyMappingPrefix + p.Mapped.Name, p.Source.Size);
                    else
                        mappedParameter = new ParameterMap(parameterPrefix + p.Mapped.Name, propertyMappingPrefix + p.Mapped.Name, p.Mapped.Properties[0].Source.DotNetType,
                                   IsOutputField(p), (p.IsNullable && p.DotNetType != "System.String"), propertyMappingPrefix + p.Mapped.Name, p.Source.Size);
                    serviceMethod.StoredProcedureMapping.Add(mappedParameter);
                }
            }
        }
        private bool IsOutputField(Property p)
        {
            if (p.IsIdentity || this.Parent.Domain.BaseEntityClass.HasRole(AttributeRoleEnum.ConcurrencyKey, p))
                return true;

            //  Now check for a derived class who's base class is an identity
            if (p.Parent.HasBaseClass && p.InPrimaryKey)
                return p.Parent.RootEntity.PrimaryKey.Properties[0].IsIdentity;

            return false;
        }
        private ServiceMethodRelationship ExposedEntity_AddRelatedMethod(Relationship source, ServiceFeatureEnum serviceFeature, ServiceMethod serviceMethod)
        {
            string paramterMappingName = "entity" + (source.RelationshipType != RelationshipTypeEnum.OneToOne ? ("." + source.Name) : "");
            string dotNetType = source.RelatedEntity.RelativeName;
            object mappingSource = null;

            if (source.RelationshipType == RelationshipTypeEnum.ManyToMany)
                mappingSource = source;
            else
                mappingSource = source.RelatedEntity;

            return this.ExposedEntity_AddRelatedMethod(mappingSource, serviceFeature, source.RelationshipType, serviceMethod);
        }
        private ServiceMethodRelationship ExposedEntity_AddRelatedMethod(object source, ServiceFeatureEnum serviceFeature, RelationshipTypeEnum relationshipType, ServiceMethod serviceMethod)
        {
            string serviceMethodName;
            BusinessService relatedBusinessService;
            string serviceName;
            ServiceMethodRelationship relatedMethod;

            serviceMethodName = this.Parent.MetaData.ServiceMethods(source, serviceFeature).MethodName;
            string ignoreServiceMethod = "";
            if (relationshipType == RelationshipTypeEnum.ManyToMany && serviceFeature == ServiceFeatureEnum.StoreSingle)
                // For M:M relationships we dont want to call back onto ourselves
                ignoreServiceMethod = serviceMethod.Parent.Name;

            relatedBusinessService = GetBusinessServiceByMethodName(serviceMethodName, ignoreServiceMethod);
            if (serviceFeature == ServiceFeatureEnum.StoreSingle && (relationshipType == RelationshipTypeEnum.OneToOne))
                serviceMethodName += "Relationships";

            serviceName = DomainModeler.Model.FixNamespace("sl." + relatedBusinessService.RelativeNamespace + "." + relatedBusinessService.Name);
            relatedMethod = new ServiceMethodRelationship(serviceName, serviceMethodName, relationshipType);

            return relatedMethod;
        }

        private void ExposedEntity_AddRelatedMethodParameterMapping(ServiceMethodRelationship relatedMethod, string relativeNamespace, string dotNetType, string parameterMappingName, ServiceMethod serviceMethod)
        {
            relatedMethod.ParameterMapping.Add(new ServiceMethodParameter(parameterMappingName, relativeNamespace, dotNetType));
            serviceMethod.RelatedMethods.Add(relatedMethod);
        }

        #endregion

        #region .   ExposedEntity - Select.Multiple.Routine

        private void CreateExposedEntity_SelectMultipleRoutine(ExposedEntityDefinition exposedEntityDefinition)
        {


            // Step 1: Create the service method
            ServiceMethod serviceMethod = this.ExposedEntity_CreateServiceMethod(ServiceFeatureEnum.RetrieveMultiple, ServiceMethodTypeEnum.RetrieveSearchResult, exposedEntityDefinition, true);

            // Step 2: Create Service Method Parameters
            serviceMethod.Parameters.Add(new ServiceMethodParameter("searchCriteria", "", serviceMethod.SourceEntity.Name + "SearchCriteria"));

            if (exposedEntityDefinition.ReturnType != null)
                serviceMethod.ReturnType = new ServiceReturnType(this.Parent.Model, exposedEntityDefinition.ReturnType.IsEnumerable,
                    exposedEntityDefinition.ReturnType.IsKnownType, exposedEntityDefinition.ReturnType.Namespace, exposedEntityDefinition.ReturnType.Type);

            // Step 3: Map Service method Parameters to Stored Procedure parameters

            // This is a specialized procedure which has no input that is specific to a given entity
            // as such this step is being omitted

            // Step 4: Add Related methods (this is handled by the ExposedEntity_ProcessRelatedMethods method
            //         This must be done after all service methods have been defined as we map
            //         back to these methods.

            // This step is not required as this method doesn't have any related methods.

            // Add the service method to the service definition
            this.Methods.Add(serviceMethod);
        }

        #endregion
        #endregion

        /// <summary>
        /// This routine will locate the service that a particular service method signature belongs to. Should there be two (highly unlikely)
        /// then the first occurance will be taken. This routine MUST be run after all business services have been defined. As such relationships
        /// are generally run as a second pass.
        /// </summary>
        /// <param name="methodName"></param>
        /// <param name="ignoreServiceMethod">Used to ignore a service method which may have the method name, which is known to be incorrect. This is typically useful on M:M relationships where the related service has the same method name as the parent service.</param>
        /// <returns></returns>
        private BusinessService GetBusinessServiceByMethodName(string methodName, string ignoreServiceMethod)
        {
            foreach (BusinessService bs in this.Parent.BusinessServices)
            {
                if (bs.Name != ignoreServiceMethod)
                    foreach (ServiceMethod sm in bs.Methods)
                        if (sm.Name == methodName)
                            return bs;


            }
            // if you get here then no match was found, this time we'll do the same thing but check any service method
            foreach (BusinessService bs in this.Parent.BusinessServices)
                foreach (ServiceMethod sm in bs.Methods)
                    if (sm.Name == methodName)
                        return bs;

            throw new ArgumentException("Unable to locate a service with the method name: " + methodName);
        }
        public BusinessLayer Parent
        {
            get { return _parent; }
        }
        public dnpModel Model
        {
            get { return this.Parent.Model; }
        }
        public string Name
        {
            get { return _name; }
        }

        public ServiceMethodCollection Methods
        {
            get { return _methods; }
        }

        #region Namespace Definitions

        public string RelativeNamespace
        {
            get { return DomainModeler.Model.FixNamespace(Naming.EvaluateNamingExpression(_definition.RelativeNamespace, this)); }
        }

        public string Namespace
        {
            get { return DomainModeler.Model.FixNamespace(this.Model.BusinessLayer.ServiceLayerNamespace + "." + this.RelativeNamespace); }
        }

        public string GeneratedFilename
        {
            get { return this.Parent.MetaData.Service.GeneratedFilename(this); }
        }

        public string Filename
        {
            get { return this.Parent.MetaData.Service.Filename(this); }
        }

        public string InterfaceGeneratedFilename
        {
            get { return this.Parent.MetaData.Service.InterfaceGeneratedFileame(this); }
        }

        public string InterfaceFilename
        {
            get { return this.Parent.MetaData.Service.InterfaceFilename(this); }
        }

        public string InterfaceRelativeNamespace
        {
            get { return DomainModeler.Model.FixNamespace(this.Parent.MetaData.Service.InterfaceRelativeNamespace(this)); }
        }

        public string InterfaceNamespace
        {
            get { return DomainModeler.Model.FixNamespace(this.Parent.Model.BusinessLayer.Model.RootNamespace + "." + this.InterfaceRelativeNamespace); }
        }
        
        public string InterfaceName
        {
            get { return this.Parent.MetaData.Service.InterfaceName(this); }
        }

        public string UnitTestFilename
        {
            get { return this.Parent.MetaData.UnitTest.Filename(this); }
        }

        public string UnitTestRelativeNamespace
        {
            get { return DomainModeler.Model.FixNamespace(this.Parent.MetaData.UnitTest.UnitTestRelativeNamespace(this)); }
        }

        public string UnitTestNamespace
        {
            get { return DomainModeler.Model.FixNamespace(this.Parent.Model.BusinessLayer.Model.RootNamespace + "." + this.UnitTestRelativeNamespace); }
        }

    #endregion
    }
}
