#region License
// (c) Garry McGlennon 2009
// 
// This project is made available under the Microsoft Public License (Ms-PL)
// Code is provided as is and with no warrenty  Use at your own risk
// 
// View the project and the latest code at http://dnpframework.codeplex.com/
// 
// All other rights reserved.
// 
// blog: www.dotNetProfessional.com/blog/
#endregion
using System;
using dnp.Framework.Database.DatabaseExplorer;
using dnp.DomainModeler;
using Relationship = dnp.DomainModeler.Relationship;

namespace dnp.ServiceModel
{
    public class DomainExtension
    {
        public static PropertyExCollection GetEntityView(Entity entity, EntityViewEnum view, PropertyDepthFilterEnum propertyFilter)
        {
            if (view == EntityViewEnum.Entity)
                return GetEntityView(entity, propertyFilter, new PropertyExCollection());

            if (view == EntityViewEnum.Query)
                return GetQueryView(entity, propertyFilter, new PropertyExCollection());
            return null;
        }

        private static PropertyExCollection GetQueryView(Entity entity, PropertyDepthFilterEnum propertyFilter, PropertyExCollection processedProperties)
        {
            PropertyExCollection propertyView = new PropertyExCollection();

            // Determine if the entity is what the user wants
            if (PropertyDepthFilterEnum.Entity == (PropertyDepthFilterEnum.Entity & propertyFilter))
            {
                // Add any other properties that are general columns.
                foreach (Property p in entity.Properties(PropertyFilterEnum.Field | PropertyFilterEnum.PrimaryKey))
                    // Check that the field we're about to add hasn't been added already
                    if (!processedProperties.Contains(p.Source.Name))
                        if (!p.InRelationshipType(RelationshipTypeEnum.ManyToOne))
                            propertyView.Add(new PropertyEx(p, p, p.Name));

                // Define our genuine mapped properties
                propertyView.AddRange(GetMappedProperties(entity, PropertyDepthFilterEnum.Entity, MappedPropertyFilterEnum.All));
            }

            // Now process the base class if requested
            if (PropertyDepthFilterEnum.BaseClass == (PropertyDepthFilterEnum.BaseClass & propertyFilter))
                if (entity.HasBaseClass)
                    // Add all the base class properties
                    propertyView.AddRange(GetQueryView(entity.BaseClass, PropertyDepthFilterEnum.BaseClass | PropertyDepthFilterEnum.Entity, propertyView));

            // Now process the base class if requested
            if (PropertyDepthFilterEnum.DerivedClass == (PropertyDepthFilterEnum.DerivedClass & propertyFilter))
                if (entity.DerivedClasses.Count != 0)
                    // Add any derived classes to the property list too!
                    foreach (Entity d in entity.DerivedClasses)
                        propertyView.AddRange(GetQueryView(d, PropertyDepthFilterEnum.DerivedClass | PropertyDepthFilterEnum.Entity, propertyView));

            return propertyView;
        }

        private static PropertyExCollection GetEntityView(Entity entity, PropertyDepthFilterEnum propertyFilter, PropertyExCollection processedProperties)
        {
            PropertyExCollection propertyView = new PropertyExCollection();

            // Determine if the entity is what the user wants
            if (PropertyDepthFilterEnum.Entity == (PropertyDepthFilterEnum.Entity & propertyFilter))
                // Define primary keys that are not part of the base entity properties
                if (entity.HasPrimaryKey && processedProperties.Count == 0)
                    foreach (Property p in entity.PrimaryKey.Properties)
                        if (!p.InRelationshipType(RelationshipTypeEnum.ManyToOne))
                            propertyView.Add(new PropertyEx(p, p, p.Name));

            // Add any other properties that are general columns.
            foreach (Property p in entity.Properties(PropertyFilterEnum.Field))
                // Check that the field we're about to add hasn't been added already
                if (!processedProperties.Contains(p.Source.Name))
                    propertyView.Add(new PropertyEx(p, p, p.Name));

            // Define our genuine mapped properties
            propertyView.AddRange(GetMappedProperties(entity, PropertyDepthFilterEnum.Entity, MappedPropertyFilterEnum.All));

            // Now process the base class if requested
            if (PropertyDepthFilterEnum.BaseClass == (PropertyDepthFilterEnum.BaseClass & propertyFilter))
                if (entity.HasBaseClass)
                    // Add all the base class properties
                    propertyView.AddRange(GetEntityView(entity.BaseClass, PropertyDepthFilterEnum.BaseClass | PropertyDepthFilterEnum.Entity, propertyView));


            // Now process the base class if requested
            if (PropertyDepthFilterEnum.DerivedClass == (PropertyDepthFilterEnum.DerivedClass & propertyFilter))
                if (entity.DerivedClasses.Count != 0)
                    // Add any derived classes to the property list too!
                    foreach (Entity d in entity.DerivedClasses)
                        propertyView.AddRange(GetEntityView(d, PropertyDepthFilterEnum.DerivedClass | PropertyDepthFilterEnum.Entity, propertyView));

            return propertyView;
        }

        public static PropertyExCollection GetMappedProperties(Entity entity, PropertyDepthFilterEnum propertyFilter, MappedPropertyFilterEnum filter)
        {
            PropertyExCollection properties = new PropertyExCollection();

            if (filter == MappedPropertyFilterEnum.NonMapped || filter == MappedPropertyFilterEnum.All)
                // Step 1: Add Many to One & Recursive Relationships that can't be mapped
                foreach (Property p in ManyToOneNonMappedProperties(entity, propertyFilter))
                    if (!properties.Contains(p.Source.Name))
                        properties.Add(new PropertyEx(p, p, p.Name));

            if (filter == MappedPropertyFilterEnum.Mapped || filter == MappedPropertyFilterEnum.All)
                // Step 2: Add Many to One Relationships that CAN be mapped
                properties.AddRange(ManyToOneMappedProperties(entity, propertyFilter));

            return properties;
        }

        private static PropertyExCollection ManyToOneMappedProperties(Entity entity, PropertyDepthFilterEnum propertyFilter)
        {
            PropertyExCollection properties = new PropertyExCollection();

            foreach (Relationship r in entity.Relationships(RelationshipTypeEnum.Recursive | RelationshipTypeEnum.ManyToOne, propertyFilter))
                if (IsMappable(r))
                {
                    Property mappedProperty = null;
                    if (r.RelatedEntity.Properties(VisibilityEnum.Visible, PropertyDepthFilterEnum.BaseClass).Contains(r.RelatedEntity.BusinessKey[0].Name))
                        mappedProperty = r.RelatedEntity.Properties(VisibilityEnum.Visible, PropertyDepthFilterEnum.BaseClass)[r.RelatedEntity.BusinessKey[0].Name];
                    else
                    {
                        // Find the property in the relationships
                        foreach (Relationship relatedRelationship in r.RelatedEntity.Relationships(RelationshipTypeEnum.ManyToOne, PropertyDepthFilterEnum.BaseClass))
                        {
                            foreach (Property p in relatedRelationship.Properties)
                                if (p.Source.Name == r.RelatedEntity.BusinessKey[0].Name)
                                {
                                    mappedProperty = p;
                                    break;
                                }
                            // Only continue if we've not found the property
                            if (mappedProperty != null)
                                break;
                        }
                    }
                    if (mappedProperty == null)
                        throw new ArgumentException(string.Format("Unable to map property for entity {0}", entity.Name));

                    string relationshipName = r.Name;
                    if (r.RelationshipType == RelationshipTypeEnum.Recursive)
                        relationshipName = r.Properties[0].Name;

                    properties.Add(new PropertyEx(r.Properties[0], mappedProperty, relationshipName));
                }
            return properties;
        }

        private static PropertyCollection ManyToOneNonMappedProperties(Entity entity, PropertyDepthFilterEnum propertyFilter)
        {
            PropertyCollection properties = new PropertyCollection();

            foreach (Relationship r in entity.Relationships(RelationshipTypeEnum.Recursive | RelationshipTypeEnum.ManyToOne, propertyFilter))
                if (!IsMappable(r))
                    // We could have multiple properties
                    foreach (Property p in r.Properties)
                        properties.Add(p);

            return properties;
        }

        private static bool IsMappable(Relationship r)
        {
            if (r.Properties.Count != 1 || r.RelatedEntity.BusinessKey.Count != 1)
                // Only single property relationships can be mapped at this time
                return false;
            else
            {
                if (r.RelatedEntity.BusinessKey[0].Name == r.RelatedEntity.PrimaryKey.Properties[0].Name)
                    return false;
                else
                    return true;
            }
        }
    }
}
